/*******************************************************************\
| MEA_Clnt.CC                                                       |
+-------------------------------------------------------------------+
| Core measurement routines for the client side.                    |
+-------------------------------------------------------------------+
| Copyright (C) Petr Tuma <petr.tuma@mff.cuni.cz>                   |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include "com_defs.h"
#include "orb_defs.h"
#include "thr_defs.h"

#include "shr_defs.h"
#include "all_defs.h"

//-------------------------------------------------------------------

#define MEA_MaxLineLen          1024

// MEA_MaxThreads               Maximum number of parallel threads

#define MEA_MaxThreads	        256

// The members of the tArgs structure are volatile because they
// can be accessed from multiple threads in parallel. Everything
// should be volatile in fact, but that does seem an overkill for
// contemporary compilers.

struct tArgsMeasure
{
  // Measured action

  MEAAction                     *pAction;
  tMEAStamp                     *asStamps;
  uint32                        iStamps;

  // Shared data for threads

  THRLock                       sLock;
  volatile uint32               iThreadsToRun;
  volatile uint32               iThreadPrincipal;

  #if   defined DSPFormatRAW

    // Shared data for measurement loop

    THRLock                       sDoRawEnd;
    volatile int                  iDoRawEnd;

    // Shared data for measurement results

    uint64                        aaiTimeStamps [MEA_MaxStamps] [MEA_RawCycles];
    uint64                        aiTimeRoundtrip [MEA_RawCycles];

  #elif defined DSPFormatTXT

    // Shared data for measurement loop

    THRLock                       sDoThreadBeg;
    THRLock                       sDoSoloEnd;
    volatile int                  iDoSoloBeg;
    volatile int                  iDoSharedBeg;
    volatile int                  iDoSharedEnd;

    // Shared data for measurement results

    uint64                        aaiTimeStamps [MEA_MaxStamps] [MEA_SoloCycles];
    uint64                        aiTimeRoundtrip [MEA_SoloCycles];
    uint64                        aiTotalCounts [MEA_MaxThreads];
    tMEAMinAvgMax<uint64>         aoTimeStamps [MEA_MaxStamps];
    tMEAMinAvgMax<uint64>         oTimeRoundtrip;

  #else
    #error
  #endif
};

//-------------------------------------------------------------------

/*******************************************************************\
| ThreadMeasure                                                     |
\*******************************************************************/

void *ThreadMeasure (void *pArgs)
{
  tArgsMeasure                  &sArgsMeasure = *(tArgsMeasure *) pArgs;
  tMEAArgs                      sArgsAction;
  uint64                        iTimeBefore, iTimeAfter;
  uint32                        iThreadNumber;
  uint32                        iMeasurement;
  uint32                        iStamp;

  ORBEnvVar
  ORBExcTry
  {
    // Determine the number of the current thread and decrement
    // the number of threads to run in the benchmark session.

    THRLockLock (&sArgsMeasure.sLock);
    sArgsMeasure.iThreadsToRun --;
    iThreadNumber = sArgsMeasure.iThreadsToRun;
    THRLockUnlock (&sArgsMeasure.sLock);

    // The exact version of the measurement loop depends on output format.


    #if   defined DSPFormatRAW

      // RAW format of measurement output.

      // It is only the principal thread that collects data in the
      // solo phase of the benchmark, because otherwise it would
      // simply be too much data.

      if (iThreadNumber == sArgsMeasure.iThreadPrincipal)
      {
        for (iMeasurement = 0 ;
             iMeasurement < MEA_RawCycles ;
             iMeasurement ++)
        {
          iTimeBefore = TIMGetTimestamp ();
          (*sArgsMeasure.pAction) (&sArgsAction ORBArgNext); ORBExcCheck
          iTimeAfter = TIMGetTimestamp ();
          for (iStamp = 0 ; iStamp < sArgsMeasure.iStamps ; iStamp ++)
          {
            sArgsMeasure.aaiTimeStamps [iStamp] [iMeasurement] =
              (sArgsMeasure.asStamps [iStamp].iFlags & MEA_StampAbsolute) ?
                (sArgsAction.aiTimes [iStamp] - iTimeBefore) :
                (sArgsAction.aiTimes [iStamp]);
          }
          sArgsMeasure.aiTimeRoundtrip [iMeasurement] = iTimeAfter - iTimeBefore;
        }

        THRLockUnlock (&sArgsMeasure.sDoRawEnd);
      }

      // All threads run until they are told to stop.

      while (!sArgsMeasure.iDoRawEnd)
      {
        (*sArgsMeasure.pAction) (&sArgsAction ORBArgNext); ORBExcCheck
      }


    #elif defined DSPFormatTXT

      // TXT format of measurement output.

      // The last thread to execute will notify the coordination
      // routine so that it can start the measurement and all.

      if (iThreadNumber == 0) THRLockUnlock (&sArgsMeasure.sDoThreadBeg);

      // It is only the principal thread that collects data in the
      // solo phase of the benchmark, because otherwise it would
      // simply be too much data.

      if (iThreadNumber == sArgsMeasure.iThreadPrincipal)
      {
        while (!sArgsMeasure.iDoSoloBeg)
        {
          (*sArgsMeasure.pAction) (&sArgsAction ORBArgNext); ORBExcCheck
        }

        for (iMeasurement = 0 ;
             iMeasurement < MEA_SoloCycles ;
             iMeasurement ++)
        {
          iTimeBefore = TIMGetTimestamp ();
          (*sArgsMeasure.pAction) (&sArgsAction ORBArgNext); ORBExcCheck
          iTimeAfter = TIMGetTimestamp ();
          for (iStamp = 0 ; iStamp < sArgsMeasure.iStamps ; iStamp ++)
          {
            sArgsMeasure.aaiTimeStamps [iStamp] [iMeasurement] =
              (sArgsMeasure.asStamps [iStamp].iFlags & MEA_StampAbsolute) ?
                (sArgsAction.aiTimes [iStamp] - iTimeBefore) :
                (sArgsAction.aiTimes [iStamp]);
          }
          sArgsMeasure.aiTimeRoundtrip [iMeasurement] = iTimeAfter - iTimeBefore;
        }

        THRLockUnlock (&sArgsMeasure.sDoSoloEnd);
      }

      // All threads collect timing information until they are
      // told to stop. The information is then added together.

      tMEAMinAvgMax<uint64> aoTimeStamps [MEA_MaxStamps];
      tMEAMinAvgMax<uint64> oTimeRoundtrip;

      for (iStamp = 0 ; iStamp < sArgsMeasure.iStamps ; iStamp ++)
      {
        aoTimeStamps [iStamp].Initialize ();
      }
      oTimeRoundtrip.Initialize ();
      sArgsMeasure.aiTotalCounts [iThreadNumber] = 0;

      while (!sArgsMeasure.iDoSharedBeg)
      {
        (*sArgsMeasure.pAction) (&sArgsAction ORBArgNext); ORBExcCheck
      }

      while (!sArgsMeasure.iDoSharedEnd)
      {
        iTimeBefore = TIMGetTimestamp ();
        (*sArgsMeasure.pAction) (&sArgsAction ORBArgNext); ORBExcCheck
        iTimeAfter = TIMGetTimestamp ();
        for (iStamp = 0 ; iStamp < sArgsMeasure.iStamps ; iStamp ++)
        {
          aoTimeStamps [iStamp].Update (
            (sArgsMeasure.asStamps [iStamp].iFlags & MEA_StampAbsolute) ?
              (sArgsAction.aiTimes [iStamp] - iTimeBefore) :
              (sArgsAction.aiTimes [iStamp]));
        }
        oTimeRoundtrip.Update (iTimeAfter - iTimeBefore);
        sArgsMeasure.aiTotalCounts [iThreadNumber] ++;
      }

      THRLockLock (&sArgsMeasure.sLock);
      for (iStamp = 0 ; iStamp < sArgsMeasure.iStamps ; iStamp ++)
      {
        sArgsMeasure.aoTimeStamps [iStamp].Update (
          aoTimeStamps [iStamp]);
      }
      sArgsMeasure.oTimeRoundtrip.Update (oTimeRoundtrip);
      THRLockUnlock (&sArgsMeasure.sLock);


    #else

      // Undefined format of measurement output.

      #error

    #endif

  }
  ORBExcCatch
  {
    DSPFatal ("MEA", "Thread " << ORBExcShow)
  }

  return (NULL);
}

//-------------------------------------------------------------------

/*******************************************************************\
| MEAMeasure                                                        |
|-------------------------------------------------------------------|
| -I- pName        Name of the action to be measured                |
|     pAction      Action to be measured                            |
|     asStamps     Stamps that the action returns                   |
|     iStamps      Number of stamps that the action returns         |
|     iThreads     Number of threads to start                       |
|     iSamples     Number of samples to make                        |
\*******************************************************************/

void MEAMeasure (char *pName,
                 MEAAction *pAction,
                 tMEAStamp *asStamps, uint32 iStamps,
                 uint32 iThreads,
                 uint32 iSamples
                 ORBDclNext) ORBDclThrow
{
  char                  acLine [MEA_MaxLineLen];
  CORBA::String_var     vLine;

  THRThread             asThreads [MEA_MaxThreads];
  CORBA::Boolean        iThreadUsage;
  tArgsMeasure          sArgs;

  uint64                iTimeHalfway, iTimerScale, iTimerGranularity;
  uint64                iTimeBefore, iTimeAfter, iTimeDelta;
  uint64                iTotalCount;

  uint32                iThread, iSample, iMeasurement, iStamp;

  assert (iThreads <= MEA_MaxThreads);

  ORBExcTry
  {
    // Print the measurement header on both client and server.

    #if   defined DSPFormatRAW

    #elif defined DSPFormatTXT

      snprintf (acLine, sizeof (acLine),
                DSP_HeaderMeasure "[1]\tSamples\t%lu\t%s\n", iSamples, pName);
      vLine = CORBA::string_dup (acLine);
      DSPStdOut << vLine << DSPFlush;
      pMEAServer->PrintString (vLine.in () ORBArgNext); ORBExcCheck

    #else
      #error
    #endif

    // Request for recording server time will be overriden when
    // the server clock is out of sync with the local clock.

    iTimeBefore = TIMGetTimestamp ();
    pMEAServer->TimerQuery (iTimeHalfway, iTimerScale, iTimerGranularity ORBArgNext); ORBExcCheck
    iTimeAfter = TIMGetTimestamp ();

    if ((iTimerScale != iTIMTimerScale) ||
        (iTimeBefore > iTimeHalfway) ||
        (iTimeAfter < iTimeHalfway))
    {
      for (iStamp = 0 ; iStamp < iStamps ; iStamp ++)
      {
        asStamps [iStamp].iFlags &= ~MEA_StampRemote;
      }
    }

    // Initialize those parts of the shared thread arguments
    // that do not change between individual samples.

    sArgs.pAction = pAction;
    sArgs.asStamps = asStamps;
    sArgs.iStamps = iStamps;

    THRLockCreate (&sArgs.sLock);

    #if   defined DSPFormatRAW

      THRLockCreate (&sArgs.sDoRawEnd); THRLockLock (&sArgs.sDoRawEnd);

    #elif defined DSPFormatTXT

      THRLockCreate (&sArgs.sDoThreadBeg); THRLockLock (&sArgs.sDoThreadBeg);
      THRLockCreate (&sArgs.sDoSoloEnd); THRLockLock (&sArgs.sDoSoloEnd);

    #else
      #error
    #endif

    for (iSample = 0 ; iSample < iSamples ; iSample ++)
    {
      // Print the sample header on both client and server.

      #if   defined DSPFormatRAW

      #elif defined DSPFormatTXT

        snprintf (acLine, sizeof (acLine),
                  DSP_HeaderSample);
        vLine = CORBA::string_dup (acLine);
        DSPStdOut << vLine << DSPFlush;
        pMEAServer->PrintString (vLine.in () ORBArgNext); ORBExcCheck

      #else
        #error
      #endif

      // Launch threads that will do the measurement.

      sArgs.iThreadsToRun         = iThreads;
      sArgs.iThreadPrincipal      = RNDGetRandom (iThreads);

      #if   defined DSPFormatRAW

        sArgs.iDoRawEnd             = FALSE;

      #elif defined DSPFormatTXT

        sArgs.iDoSoloBeg            = FALSE;
        sArgs.iDoSharedBeg          = FALSE;
        sArgs.iDoSharedEnd          = FALSE;

        for (iStamp = 0 ; iStamp < iStamps ; iStamp ++)
        {
          sArgs.aoTimeStamps [iStamp].Initialize ();
        }
        sArgs.oTimeRoundtrip.Initialize ();

      #else
        #error
      #endif

      for (iThread = 0 ; iThread < iThreads ; iThread ++)
        THRThreadCreate (&asThreads [iThread], ThreadMeasure, &sArgs);

      // The exact progress of the measurement loop depends on output format.


      #if   defined DSPFormatRAW

        // PHASE RAW
        // Measures everything without precautions.

        THRLockLock (&sArgs.sDoRawEnd);
        sArgs.iDoRawEnd = TRUE;


      #elif defined DSPFormatTXT

        // PHASE THREAD
        // Waits for all threads to start.

        THRLockLock (&sArgs.sDoThreadBeg);

        // PHASE WARMUP
        // Waits for all threads to become warmed up.

        THRThreadSleep (MEA_WarmUpTimeMin +
                        RNDGetRandom (MEA_WarmUpTimeMax -
                                      MEA_WarmUpTimeMin));

        // PHASE USAGE CLIENT
        // Measures resource usage on the client.

        MEAUsage ();

        // PHASE USAGE SERVER
        // Measures resource usage on the server.

        pMEAServer->UsageStart (ORBArgOnly); ORBExcCheck
        THRThreadSleep (MEA_UsageCycles);
        do
        {
          THRThreadSleep (1);
          iThreadUsage = pMEAServer->UsageQuery (ORBArgOnly); ORBExcCheck
        }
        while (iThreadUsage);

        // PHASE SOLO
        // Measures individual samples from one thread.

        sArgs.iDoSoloBeg = TRUE;
        THRLockLock (&sArgs.sDoSoloEnd);

        // PHASE SHARED
        // Measures averaged samples from all threads.

        iTimeBefore = TIMGetTimestamp ();
        sArgs.iDoSharedBeg = TRUE;
        THRThreadSleep (MEA_SharedTime);
        iTimeAfter = TIMGetTimestamp ();
        sArgs.iDoSharedEnd = TRUE;
        iTimeDelta = iTimeAfter - iTimeBefore;


      #else
        #error
      #endif

      // Wait for the threads to finish before doing anything.

      for (iThread = 0 ; iThread < iThreads ; iThread ++)
        THRThreadFinish (&asThreads [iThread]);

      // Display the rest of the output depending on the output format.


      #if   defined DSPFormatRAW

        for (iMeasurement = 0 ; iMeasurement < MEA_RawCycles ; iMeasurement ++)
        {
          for (iStamp = 0 ; iStamp < iStamps ; iStamp ++)
          {
            if (asStamps [iStamp].iFlags & (MEA_StampLocal | MEA_StampRemote))
            {
              DSPStdOut << sArgs.aaiTimeStamps [iStamp] [iMeasurement] << '\t';
            }
          }
          DSPStdOut << sArgs.aiTimeRoundtrip [iMeasurement] << '\t' << DSPEndLine;
        }


      #elif defined DSPFormatTXT

        DSPStdOut << "[" << iThreads << "]\tThruputSolo\t";
        iTotalCount = 0;
        for (iThread = 0 ; iThread < iThreads ; iThread ++)
        {
          DSPStdOut << (sArgs.aiTotalCounts [iThread] * iTIMTimerScale) / iTimeDelta << '\t';
          iTotalCount += sArgs.aiTotalCounts [iThread];
        }
        DSPStdOut << "[1]\tThruputAverage\t" << (iTotalCount * iTIMTimerScale) / iTimeDelta << '\t';

        for (iStamp = 0 ; iStamp < iStamps ; iStamp ++)
        {
          if (asStamps [iStamp].iFlags & (MEA_StampLocal | MEA_StampRemote))
          {
            sArgs.aoTimeStamps [iStamp].Normalize (iTotalCount);
            snprintf (acLine, sizeof (acLine), "%sAverage", asStamps [iStamp].pName);
            sArgs.aoTimeStamps [iStamp].Print (acLine);

            DSPStdOut << "[" << MEA_SoloCycles << "]\t" << asStamps [iStamp].pName << "Solo\t";
            for (iMeasurement = 0 ; iMeasurement < MEA_SoloCycles ; iMeasurement ++)
            {
              DSPStdOut << sArgs.aaiTimeStamps [iStamp] [iMeasurement] << '\t';
            }
          }
        }

        sArgs.oTimeRoundtrip.Normalize (iTotalCount);
        sArgs.oTimeRoundtrip.Print ("RoundtripAverage");

        DSPStdOut << "[" << MEA_SoloCycles << "]\tRoundtripSolo\t";
        for (iMeasurement = 0 ; iMeasurement < MEA_SoloCycles ; iMeasurement ++)
        {
          DSPStdOut << sArgs.aiTimeRoundtrip [iMeasurement] << '\t';
        }
        snprintf (acLine, sizeof (acLine), "\n");
        vLine = CORBA::string_dup (acLine);
        DSPStdOut << vLine << DSPFlush;
        pMEAServer->PrintString (vLine.in () ORBArgNext); ORBExcCheck


      #else
        #error
      #endif
    }

    THRLockFinish (&sArgs.sLock);

    #if   defined DSPFormatRAW

      THRLockUnlock (&sArgs.sDoRawEnd); THRLockFinish (&sArgs.sDoRawEnd);

    #elif defined DSPFormatTXT

      THRLockUnlock (&sArgs.sDoThreadBeg); THRLockFinish (&sArgs.sDoThreadBeg);
      THRLockUnlock (&sArgs.sDoSoloEnd); THRLockFinish (&sArgs.sDoSoloEnd);

    #else
      #error
    #endif
  }
  ORBExcCatch
  {
    DSPFatal ("MEA", "Thread " << ORBExcShow)
  }
}

//-------------------------------------------------------------------

