/*******************************************************************\
| MEA_Srvr.CC                                                       |
+-------------------------------------------------------------------+
| Core measurement routines for the server side.                    |
+-------------------------------------------------------------------+
| Copyright (C) Petr Tuma <petr.tuma@mff.cuni.cz>                   |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include "com_defs.h"
#include "orb_defs.h"
#include "thr_defs.h"

#include "shr_defs.h"
#include "all_defs.h"

//-------------------------------------------------------------------

/*******************************************************************\
| MEAUsage                                                          |
\*******************************************************************/

void MEAUsage (void)
{
  // Variables related to memory usage.

  tMEAMinAvgMax<uint64>         oMemoryApplicationResident;
  tMEAMinAvgMax<uint64>         oMemoryApplicationSwapped;
  tMEAMinAvgMax<uint64>         oMemoryKernelUsed;
  tMEAMinAvgMax<uint64>         oMemoryTotalPhysicalUsed;
  tMEAMinAvgMax<uint64>         oMemoryTotalPhysicalFree;
  tMEAMinAvgMax<uint64>         oMemoryTotalSwapUsed;

  uint64     iMemoryApplicationResident;
  uint64     iMemoryApplicationSwapped;
  uint64     iMemoryKernelUsed;
  uint64     iMemoryTotalPhysicalUsed;
  uint64     iMemoryTotalPhysicalFree;
  uint64     iMemoryTotalSwapUsed;

  // Variables related to network usage.

  tMEAMinAvgMax<uint64>         oNetworkBytesSent;
  tMEAMinAvgMax<uint64>         oNetworkBytesReceived;
  tMEAMinAvgMax<uint64>         oNetworkPacketsSent;
  tMEAMinAvgMax<uint64>         oNetworkPacketsReceived;

  uint64     iNetworkBytesSentBefore,       iNetworkBytesSentAfter;
  uint64     iNetworkBytesReceivedBefore,   iNetworkBytesReceivedAfter;
  uint64     iNetworkPacketsSentBefore,     iNetworkPacketsSentAfter;
  uint64     iNetworkPacketsReceivedBefore, iNetworkPacketsReceivedAfter;

  uint64     iNetworkDummy;

  // Variables related to processor usage.

  tMEAMinAvgMax<uint64>         oProcessorApplicationThreads;
  tMEAMinAvgMax<int>            oProcessorTotalKernel;
  tMEAMinAvgMax<int>            oProcessorTotalWait;
  tMEAMinAvgMax<int>            oProcessorTotalUser;
  tMEAMinAvgMax<int>            oProcessorTotalIdle;

  uint64        iProcessorApplicationThreads;

  uint64        iProcessorTotalKernelBefore,   iProcessorTotalKernelAfter;
  uint64        iProcessorTotalWaitBefore,     iProcessorTotalWaitAfter;
  uint64        iProcessorTotalUserBefore,     iProcessorTotalUserAfter;
  uint64        iProcessorTotalIdleBefore,     iProcessorTotalIdleAfter;

  uint64        iProcessorDummy, iProcessorTotal;

  // Other variables.

  uint64        iTimeBefore, iTimeAfter, iTimeDelta;
  int           iMeasurement;

  // We have a lot of values to monitor, some things are
  // therefore better done through a set of macros.

  #define ApplyOne(Name,What)                                   \
  {                                                             \
    const char *pName = # Name;                                 \
    o##Name.What;                                               \
  };

  #define ApplyAll(What)                                        \
  {                                                             \
    ApplyOne (MemoryApplicationResident, What)                  \
    ApplyOne (MemoryApplicationSwapped, What)                   \
    ApplyOne (MemoryKernelUsed, What)                           \
    ApplyOne (MemoryTotalPhysicalUsed, What)                    \
    ApplyOne (MemoryTotalPhysicalFree, What)                    \
    ApplyOne (MemoryTotalSwapUsed, What)                        \
    ApplyOne (NetworkBytesSent, What)                           \
    ApplyOne (NetworkBytesReceived, What)                       \
    ApplyOne (NetworkPacketsSent, What)                         \
    ApplyOne (NetworkPacketsReceived, What)                     \
    ApplyOne (ProcessorApplicationThreads, What)                \
    ApplyOne (ProcessorTotalKernel, What)                       \
    ApplyOne (ProcessorTotalWait, What)                         \
    ApplyOne (ProcessorTotalUser, What)                         \
    ApplyOne (ProcessorTotalIdle, What)                         \
  }

  // Initialize all values monitored.

  ApplyAll (Initialize ())

  // The snapshot of time is always taken before the snapshot
  // of other system information to compensate for the time
  // spent inside the routines that query this information.

  iTimeBefore = TIMGetTimestamp ();

  NETGetTotalTraffic (&iNetworkBytesSentBefore, &iNetworkBytesReceivedBefore, &iNetworkPacketsSentBefore, &iNetworkPacketsReceivedBefore);
  PROGetTotalUsage (&iProcessorTotalKernelBefore, &iProcessorTotalWaitBefore, &iProcessorTotalUserBefore, &iProcessorTotalIdleBefore);

  for (iMeasurement = 0 ; iMeasurement < MEA_UsageCycles ; iMeasurement ++)
  {
    THRThreadSleep (1);

    iTimeAfter = TIMGetTimestamp ();
    iTimeDelta = iTimeAfter - iTimeBefore;
    iTimeBefore = iTimeAfter;

    MEMGetApplicationMemory  (&iMemoryApplicationResident, &iMemoryApplicationSwapped);
    MEMGetKernelMemory       (&iMemoryKernelUsed);
    MEMGetTotalMemory        (&iMemoryTotalPhysicalUsed, &iMemoryTotalPhysicalFree, &iMemoryTotalSwapUsed);
    NETGetTotalTraffic       (&iNetworkBytesSentAfter, &iNetworkBytesReceivedAfter, &iNetworkPacketsSentAfter, &iNetworkPacketsReceivedAfter);
    PROGetApplicationThreads (&iProcessorApplicationThreads);
    PROGetTotalUsage         (&iProcessorTotalKernelAfter, &iProcessorTotalWaitAfter, &iProcessorTotalUserAfter, &iProcessorTotalIdleAfter);

    oMemoryApplicationResident.Update (iMemoryApplicationResident);
    oMemoryApplicationSwapped.Update  (iMemoryApplicationSwapped);
    oMemoryKernelUsed.Update          (iMemoryKernelUsed);
    oMemoryTotalPhysicalUsed.Update   (iMemoryTotalPhysicalUsed);
    oMemoryTotalPhysicalFree.Update   (iMemoryTotalPhysicalFree);
    oMemoryTotalSwapUsed.Update       (iMemoryTotalSwapUsed);

    #define UpdateNetwork(Which)                                \
    {                                                           \
      iNetworkDummy = (iNetwork##Which##After -                 \
                       iNetwork##Which##Before) *               \
                      iTIMTimerScale;                           \
      oNetwork##Which.Update (iNetworkDummy / iTimeDelta);      \
      iNetwork##Which##Before = iNetwork##Which##After;         \
    };

    UpdateNetwork (BytesSent)
    UpdateNetwork (BytesReceived)
    UpdateNetwork (PacketsSent)
    UpdateNetwork (PacketsReceived)

    #undef UpdateNetwork

    oProcessorApplicationThreads.Update (iProcessorApplicationThreads);

    iProcessorTotal = iProcessorTotalKernelAfter - iProcessorTotalKernelBefore +
                      iProcessorTotalWaitAfter - iProcessorTotalWaitBefore +
                      iProcessorTotalUserAfter - iProcessorTotalUserBefore +
                      iProcessorTotalIdleAfter - iProcessorTotalIdleBefore;

    if (iProcessorTotal != 0)
    {
      #define UpdateProcessorTotal(Which)                       \
      {                                                         \
        iProcessorDummy =                                       \
          (iProcessorTotal##Which##After -                      \
           iProcessorTotal##Which##Before) * 100;               \
        oProcessorTotal##Which.Update (                         \
          iProcessorDummy / iProcessorTotal);                   \
        iProcessorTotal##Which##Before =                        \
          iProcessorTotal##Which##After;                        \
      };

      UpdateProcessorTotal (Kernel)
      UpdateProcessorTotal (Wait)
      UpdateProcessorTotal (User)
      UpdateProcessorTotal (Idle)

      #undef UpdateProcessorTotal
    };
  };

  // Normalize all values monitored and print the values.

  ApplyAll (Normalize (MEA_UsageCycles))
  ApplyAll (Print (pName))

  #undef ApplyOne
  #undef ApplyAll
};

//-------------------------------------------------------------------

