/*******************************************************************\
| SRV_Main.CC                                                       |
+-------------------------------------------------------------------+
| Copyright (C) Petr Tuma <petr.tuma@mff.cuni.cz>                   |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include "com_defs.h"
#include "orb_defs.h"
#include "thr_defs.h"

#include "shr_defs.h"
#include "all_defs.h"

//-------------------------------------------------------------------

THRThread               sThreadUsage;
volatile int            iThreadUsage;

// Do not use CORBA var classes as global variables, it causes
// problems with proper initialization and shutdown order.

CORBA::Object_ptr       pMEAReference;
Server_ptr              pMEAServer;

//-------------------------------------------------------------------

/*******************************************************************\
| ServerServant                                                     |
\*******************************************************************/

class ServerServant : public POA_Server
{
  public:

    void              PrintString (const char *sMessage
                                   ORBDclNext) ORBDclThrow;
    void              UsageStart  (ORBDclOnly) ORBDclThrow;
    CORBA::Boolean    UsageQuery  (ORBDclOnly) ORBDclThrow;
    void              TimerQuery  (CORBA::ULongLong_out iTimestamp,
                                   CORBA::ULongLong_out iScale,
                                   CORBA::ULongLong_out iGranularity
                                   ORBDclNext) ORBDclThrow;
    CORBA::Object_ptr Reference   (ORBDclOnly) ORBDclThrow;
    void              Terminate   (ORBDclOnly) ORBDclThrow;
};

//-------------------------------------------------------------------

/*******************************************************************\
| ServerServant::PrintString                                        |
\*******************************************************************/

void ServerServant::PrintString (
  const char *sMessage
  ORBDclNext) ORBDclThrow
{
  DSPStdOut << sMessage << DSPFlush;
};

//-------------------------------------------------------------------

// We cannot afford to measure the resource usage during method
// invocation, because that can block other method invocations.
// The resource usage measurement is therefore done in a
// separate thread, whose termination is queried by
// polling, hopefully infrequently enough.

/*******************************************************************\
| ThreadUsage                                                       |
\*******************************************************************/

void *ThreadUsage (void *pArgs)
{
  MEAUsage ();
  iThreadUsage = FALSE;
  return (NULL);
};

/*******************************************************************\
| ServerServant::UsageStart                                         |
\*******************************************************************/

void ServerServant::UsageStart (
  ORBDclOnly) ORBDclThrow
{
  iThreadUsage = TRUE;
  THRThreadCreate (&sThreadUsage, ThreadUsage, NULL);
};

/*******************************************************************\
| ServerServant::UsageQuery                                         |
\*******************************************************************/

CORBA::Boolean ServerServant::UsageQuery (
  ORBDclOnly) ORBDclThrow
{
  if (!iThreadUsage) THRThreadFinish (&sThreadUsage);
  return (iThreadUsage);
};

//-------------------------------------------------------------------

/*******************************************************************\
| ServerServant::TimerQuery                                         |
\*******************************************************************/

void ServerServant::TimerQuery (
  CORBA::ULongLong_out iTimestamp,
  CORBA::ULongLong_out iScale,
  CORBA::ULongLong_out iGranularity
  ORBDclNext) ORBDclThrow
{
  iTimestamp    = TIMGetTimestamp ();
  iScale        = iTIMTimerScale;
  iGranularity  = iTIMTimerGranularity;
};

//-------------------------------------------------------------------

/*******************************************************************\
| ServerServant::Reference                                          |
\*******************************************************************/

CORBA::Object_ptr ServerServant::Reference (
  ORBDclOnly) ORBDclThrow
{
  return (CORBA::Object::_duplicate (pMEAReference));
};

/*******************************************************************\
| ServerServant::Terminate                                          |
\*******************************************************************/

void ServerServant::Terminate (
  ORBDclOnly) ORBDclThrow
{
  pORB->shutdown (FALSE ORBArgNext);
};

//-------------------------------------------------------------------

int main (int iArgC, char *apArgV[])
{
  // The initialization proceeds in a specific order. First, the
  // libraries are initialized. Second, the system information
  // routines are initialized. Third, the broker is initialized.

  TIMInit ();
  RNDInit (TIMGetTimestamp ());

  MEMInit ();
  NETInit ();
  PROInit ();

  ORBInitServer (iArgC, apArgV);

  // The benchmark is contained within other modules.

  ORBEnvVar
  ORBExcTry
  {
    #if   defined DSPFormatRAW

    #elif defined DSPFormatTXT

      // Display header

      DSPStdOut << DSP_HeaderBenchie
                << "[STR]\tVersion\t" << SAM_BenchmarkName " " SAM_GlobalVersion " " SAM_GlobalDate << '\t'
                << "[STR]\tSuite\t" << BENGetName () << '\t'
                << DSPEndLine;

      // Display configuration

      DSPStdOut << DSP_HeaderConfig
                << "[1]\tScale\t" << iTIMTimerScale << '\t'
                << "[1]\tGranularity\t" << iTIMTimerGranularity << '\t'
                << DSPEndLine;

    #else
      #error
    #endif

    // Execute measurement

    ServerServant *pMEAServerServant;
    CORBA::Object_ptr pMEAServerBase;

    pMEAServerServant = new ServerServant ();
    pMEAServerBase = pRootPOA->servant_to_reference (pMEAServerServant ORBArgNext); ORBExcCheck
    pMEAServer = Server::_narrow (pMEAServerBase);

    LOCAnnounceServer (pMEAServer);

    pMEAReference = BENServer (iArgC, apArgV);

    pRootPOAManager->activate (ORBArgOnly); ORBExcCheck
    pORB->run (ORBArgOnly); ORBExcCheck
  }
  ORBExcCatch
  {
    DSPFatal ("SRV", "Main " << ORBExcShow)
  };

  ORBDoneServer ();

  return (0);
};

//-------------------------------------------------------------------

