/*******************************************************************\
| MEM_Linux_Proc.CC                                                 |
+-------------------------------------------------------------------+
| The memory information functions for Linux.                       |
+-------------------------------------------------------------------+
| Copyright (C) Petr Tuma <petr.tuma@mff.cuni.cz>                   |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include <unistd.h>
#include <sys/sysinfo.h>

#include "com_defs.h"
#include "orb_defs.h"
#include "thr_defs.h"

#include "shr_defs.h"
#include "all_defs.h"

//-------------------------------------------------------------------

#define MEM_MaxLineLen		1024

//-------------------------------------------------------------------

static uint64 iPageSize;

//-------------------------------------------------------------------

/*******************************************************************\
| MEMInit                                                           |
\*******************************************************************/

void MEMInit (void)
{
  iPageSize = sysconf (_SC_PAGE_SIZE);
}

//-------------------------------------------------------------------

/*******************************************************************\
| MEMGetApplicationMemory                                           |
\*******************************************************************/

void MEMGetApplicationMemory (uint64 *pMemoryResident,
                              uint64 *pMemorySwapped)
{
  FILE                  *pFile;
  int                   iDummy;
  uint64                iPagesTotal;
  uint64                iPagesResident;

  pFile = fopen ("/proc/self/statm", "r");
  assert (pFile != NULL);
  iDummy = fscanf (pFile, "%qu %qu", &iPagesTotal, &iPagesResident);
  assert (iDummy == 2);
  iDummy = fclose (pFile);
  assert (!iDummy);

  (*pMemoryResident) = iPagesResident * iPageSize;
  (*pMemorySwapped) = (iPagesTotal - iPagesResident) * iPageSize;
}

/*******************************************************************\
| MEMGetKernelMemory                                                |
\*******************************************************************/

void MEMGetKernelMemory (uint64 *pMemoryUsed)
{
  FILE                  *pFile;
  char                  acLine [MEM_MaxLineLen];
  void                  *pDummy;
  int                   iDummy;
  int                   iVersionMajor;
  int                   iVersionMinor;
  uint64                iSlabSize;
  uint64                iSlabCount;
  uint64                iSlabPages;
  uint64                iTotalPages;

  pFile = fopen ("/proc/slabinfo", "r");
  assert (pFile != NULL);

  // The first line of the file tells its version.

  pDummy = fgets (acLine, sizeof (acLine), pFile);
  assert (pDummy != NULL);
  iDummy = sscanf (acLine, "slabinfo - version: %i.%i", &iVersionMajor, &iVersionMinor);
  assert (iDummy == 2);

  iTotalPages = 0;

  // Version 1.1 tells number of used pages for each slab cache directly.

  if ((iVersionMajor == 1) && (iVersionMinor == 1))
  {
    while (fgets (acLine, sizeof (acLine), pFile) != NULL)
    {
      iDummy = sscanf (acLine, "%*s %*u %*u %*u %*u %qu %*u", &iSlabPages);
      assert (iDummy == 1);
      iTotalPages += iSlabPages;
    }
  }

  // Version 2.1 tells number of pages per slab and number of slabs for each slab cache.

  if ((iVersionMajor == 2) && (iVersionMinor == 1))
  {
    pDummy = fgets (acLine, sizeof (acLine), pFile);
    assert (pDummy != NULL);

    while (fgets (acLine, sizeof (acLine), pFile) != NULL)
    {
      iDummy = sscanf (acLine, "%*s %*u %*u %*u %*u %qu : tunables %*u %*u %*u : slabdata %*u %qu %*u", &iSlabSize, &iSlabCount);
      assert (iDummy == 2);
      iTotalPages += iSlabSize * iSlabCount;
    }
  }

  iDummy = fclose (pFile);
  assert (!iDummy);

  (*pMemoryUsed) = iTotalPages * iPageSize;
}

/*******************************************************************\
| MEMGetTotalMemory                                                 |
\*******************************************************************/

void MEMGetTotalMemory (uint64 *pPhysicalUsed,
                        uint64 *pPhysicalFree,
                        uint64 *pSwapUsed)
{
  struct sysinfo        sSysInfo;
  int                   iDummy;

  iDummy = sysinfo (&sSysInfo);
  assert (!iDummy);
  (*pPhysicalUsed) = (uint64) (sSysInfo.totalram - sSysInfo.freeram) * (uint64) sSysInfo.mem_unit;
  (*pPhysicalFree) = (uint64) sSysInfo.freeram * (uint64) sSysInfo.mem_unit;
  (*pSwapUsed) = (uint64) (sSysInfo.totalswap - sSysInfo.freeswap) * (uint64) sSysInfo.mem_unit;
}

//-------------------------------------------------------------------
