/*******************************************************************\
| MEM_Solaris_Proc.CC                                               |
+-------------------------------------------------------------------+
| The memory information functions for Solaris.                     |
+-------------------------------------------------------------------+
| Copyright (C) Petr Tuma <petr.tuma@mff.cuni.cz>                   |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include <kstat.h>
#include <procfs.h>
#include <unistd.h>
#include <sys/swap.h>

#include "com_defs.h"
#include "orb_defs.h"
#include "thr_defs.h"

#include "shr_defs.h"
#include "all_defs.h"

//-------------------------------------------------------------------

#define MEM_MaxSwapLen		256
#define MEM_MaxLineLen		1024

//-------------------------------------------------------------------

typedef struct tSwapTable {
  int       iEntries;
  swapent_t asEntries [MEM_MaxSwapLen];
};


//-------------------------------------------------------------------

/*******************************************************************\
| MEMInit                                                           |
\*******************************************************************/

void MEMInit (void)
{
};

//-------------------------------------------------------------------

/*******************************************************************\
| MEMGetApplicationMemory                                           |
\*******************************************************************/

void MEMGetApplicationMemory (uint64 *pMemoryResident,
                              uint64 *pMemorySwapped)
{
  FILE          *pFile;
  psinfo_t      sPSInfo;
  int	        iDummy;

  pFile = fopen ("/proc/self/psinfo", "r");
  assert (pFile != NULL);
  iDummy = fread (&sPSInfo, sizeof (sPSInfo), 1, pFile);
  assert (iDummy == 1);
  iDummy = fclose (pFile);
  assert (iDummy == 0);

  (*pMemoryResident) = ((uint64) sPSInfo.pr_rssize) * 1024;
  (*pMemorySwapped)  = ((uint64) (sPSInfo.pr_size - sPSInfo.pr_rssize)) * 1024;
};

/*******************************************************************\
| MEMGetKernelMemory                                                |
\*******************************************************************/

void MEMGetKernelMemory (uint64 *pMemoryUsed)
{
  kstat_ctl_t   *pKStats;
  kstat_t       *pKStat;
  kstat_named_t *pKValue;
  kid_t         iKStatID;
  int           iDummy;

  // We have to do this because kstat_lookup wants char pointer
  // instead of const char pointer which it probably should.

  char acLiteralUnix        [] = "unix";
  char acLiteralSystemPages [] = "system_pages";
  char acLiteralPPKernel    [] = "pp_kernel";

  // We ignore the possibility of the kstat chain changing
  // during lookup. First, this is unlikely to happen during
  // benchmarking. Second, the kstat structure we need is likely
  // to remain in the kstat chain regardless of its other changes.

  pKStats = kstat_open ();
  assert (pKStats != NULL);

  pKStat = kstat_lookup (pKStats, acLiteralUnix, 0, acLiteralSystemPages);
  assert (pKStat != NULL);
  iKStatID = kstat_read (pKStats, pKStat, NULL);
  assert (iKStatID != -1);
  pKValue = (kstat_named_t *) kstat_data_lookup (pKStat, acLiteralPPKernel);
  assert (pKValue != NULL);
  assert (pKValue->data_type == KSTAT_DATA_UINT32);
  (*pMemoryUsed) = ((uint64) pKValue->value.ui32) *
                   ((uint64) sysconf (_SC_PAGE_SIZE));

  iDummy = kstat_close (pKStats);
  assert (!iDummy);
};

/*******************************************************************\
| MEMGetTotalMemory                                                 |
\*******************************************************************/

void MEMGetTotalMemory (uint64 *pPhysicalUsed,
                        uint64 *pPhysicalFree,
                        uint64 *pSwapUsed)
{
  tSwapTable	sSwapTable;
  char          acSwapName [MEM_MaxLineLen];
  int           iSwapSegments;
  int		iSwapSegment;
  uint64        iPageSize;
  int		iDummy;

  iPageSize         = sysconf (_SC_PAGE_SIZE);

  // Physical memory is trivial, really :-) ...

  (*pPhysicalFree) = sysconf (_SC_AVPHYS_PAGES) * iPageSize;
  (*pPhysicalUsed) = (sysconf (_SC_PHYS_PAGES) * iPageSize) - (*pPhysicalFree);

  // To collect the amount of used virtual memory, we have to
  // iterate over all swapfiles in the system. This is somewhat
  // simplified to get rid of dynamic allocation of swap names.

  // If the following assertion fails, tSwapTable is probably
  // inconsistent with system definition of swaptbl_t.

  assert (sizeof (swaptbl_t) == sizeof (swapent_t) + sizeof (int));

  iSwapSegments = swapctl (SC_GETNSWP, NULL);
  assert (iSwapSegments >= 0 && iSwapSegments < MEM_MaxSwapLen);

  sSwapTable.iEntries = iSwapSegments;
  for (iSwapSegment = 0 ; iSwapSegment < iSwapSegments ; iSwapSegment ++)
  {
    sSwapTable.asEntries [iSwapSegment].ste_path = acSwapName;
  };
  iDummy = swapctl (SC_LIST, &sSwapTable);
  assert (iDummy == iSwapSegments);

  (*pSwapUsed) = 0;

  for (iSwapSegment = 0 ; iSwapSegment < iSwapSegments ; iSwapSegment ++)
  {
    (*pSwapUsed) += (sSwapTable.asEntries [iSwapSegment].ste_pages -
                     sSwapTable.asEntries [iSwapSegment].ste_free) * iPageSize;
  };
};

//-------------------------------------------------------------------
