/*******************************************************************\
| MEM_Windows_PData.CC                                              |
+-------------------------------------------------------------------+
| The memory information functions for Windows.                     |
+-------------------------------------------------------------------+
| Copyright (C) Lubomir Bulej <bulej@nenya.ms.mff.cuni.cz>          |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include <assert.h>

#include "com_defs.h"
#include "plt_defs.h"

//-------------------------------------------------------------------

#define MEM_OBJ_Memory            TEXT("Memory")
#define MEM_CTR_PoolPagedBytes    TEXT("Pool Paged Bytes")
#define MEM_CTR_PoolNonpagedBytes TEXT("Pool Nonpaged Bytes")

#define MEM_OBJ_Process           TEXT("Process")
#define MEM_CTR_IDProcess         TEXT("ID Process")
#define MEM_CTR_WorkingSet        TEXT("Working Set")
#define MEM_CTR_PageFileBytes     TEXT("Page File Bytes")

//-------------------------------------------------------------------

static DWORD iProcessID;

static DWORD iMemoryObj;
static TCHAR acMemoryObj [PLT_IndexDigits]   = { TEXT ("") };

static DWORD iPoolPagedBytesCtr;
static DWORD iPoolNonpagedBytesCtr;

static DWORD iProcessObj;
static TCHAR acProcessObj [PLT_IndexDigits]  = { TEXT ("") };

static DWORD iIDProcessCtr;
static DWORD iWorkingSetCtr;
static DWORD iPageFileBytesCtr;

//-------------------------------------------------------------------

/*******************************************************************\
| MEMInit                                                           |
\*******************************************************************/

void MEMInit (void)
{
  PLTInitialize ();
  iProcessID = GetCurrentProcessId ();

  // Memory stats
  iMemoryObj = PLTFindPerfIndex (MEM_OBJ_Memory);
  _sntprintf (acMemoryObj, PLT_IndexDigits, PLT_IndexFormat, iMemoryObj);

  iPoolPagedBytesCtr    = PLTFindPerfIndex (MEM_CTR_PoolPagedBytes);
  iPoolNonpagedBytesCtr = PLTFindPerfIndex (MEM_CTR_PoolNonpagedBytes);

  // Process memory stats
  iProcessObj = PLTFindPerfIndex (MEM_OBJ_Process);
  _sntprintf (acProcessObj, PLT_IndexDigits, PLT_IndexFormat, iProcessObj);

  iIDProcessCtr     = PLTFindPerfIndex (MEM_CTR_IDProcess);
  iWorkingSetCtr    = PLTFindPerfIndex (MEM_CTR_WorkingSet);
  iPageFileBytesCtr = PLTFindPerfIndex (MEM_CTR_PageFileBytes);
}; // MEMInit (void)

//-------------------------------------------------------------------

/*******************************************************************\
| MEMGetApplicationMemory                                           |
\*******************************************************************/

/*
 * Oh well... At first, it seemed we could use PSAPI to get the
 * information using GetProcessMemoryInfo call, but the header file
 * for PSAPI.DLL seems to be missing in VC98.
 *
 * We could then use NtQueryInformationProcess from NTDLL.DLL, but
 * that's not officially documented and subject to change.
 *
 * So we'll have to make do with performance counters...
 */
void MEMGetApplicationMemory (uint64 *pMemoryResident,
                              uint64 *pMemorySwapped)
{
  PERF_DATA_BLOCK *           pDataBlk;
  PERF_OBJECT_TYPE *          pObjType;
  PERF_COUNTER_BLOCK *        pCntrBlk;

  PERF_COUNTER_DEFINITION *   pIDProcessCtr;
  PERF_COUNTER_DEFINITION *   pWorkingSetCtr;
  PERF_COUNTER_DEFINITION *   pPageFileBytesCtr;


  PLTAcquirePerfMutex ();

  /*
   * Query Process performace object, search by process ID for instance for
   * this process and get the Thread Count counter.
   */
  for (int iterations = 0; /* true */; iterations++) {
    assert (iterations < PLT_MaxIterations);

    pDataBlk = PLTGetPerfData (acProcessObj, &pPerfDataMem, &iPerfDataLen);
    pObjType = PLTFindPerfObject (pDataBlk, iProcessObj);
    if (pObjType != NULL) break;
  } // get the performance object

  pIDProcessCtr     = PLTFindPerfCounter (pObjType, iIDProcessCtr);
  pWorkingSetCtr    = PLTFindPerfCounter (pObjType, iWorkingSetCtr);
  pPageFileBytesCtr = PLTFindPerfCounter (pObjType, iPageFileBytesCtr);

  /*
   * There will always be at least one process, so find the
   * instance with the right PID and get its memory stats.
   */
  pCntrBlk = PLTFirstCntrBlk (pObjType);
  for (LONG iInstNum = 0; iInstNum < pObjType->NumInstances; iInstNum++) {

    // look for process ID match
    if (iProcessID == PLTGetPerfCounter (pCntrBlk, pIDProcessCtr)) {
      UINT64 iWorkingSet    = PLTGetPerfCounter (pCntrBlk, pWorkingSetCtr);
      UINT64 iPageFileBytes = PLTGetPerfCounter (pCntrBlk, pPageFileBytesCtr);

      /*
       * Pagefile usage field says how much swap space is reserved for the
       * process in system's pagefiles. For rough estimate of how much of
       * a process is swapped out, we use the difference between working set
       * size and the amount of reserved swap space.
       */
      (*pMemoryResident) = iWorkingSet;
      (*pMemorySwapped) = (iPageFileBytes > iWorkingSet) ? iPageFileBytes - iWorkingSet : 0;

      PLTReleasePerfMutex ();
      return;
    };

    pCntrBlk = PLTNextCntrBlk (pCntrBlk);
  };

  assert (FALSE);
}; // MEMGetApplicationMemory (uint64, uint64)

/*******************************************************************\
| MEMGetKernelMemory                                                |
\*******************************************************************/

void MEMGetKernelMemory (uint64 *pMemoryUsed)
{
  PERF_DATA_BLOCK *           pDataBlk;
  PERF_OBJECT_TYPE *          pObjType;
  PERF_COUNTER_BLOCK *        pCntrBlk;

  PERF_COUNTER_DEFINITION *   pPoolPagedBytesCtr;
  PERF_COUNTER_DEFINITION *   pPoolNonpagedBytesCtr;


  PLTAcquirePerfMutex ();

  /*
   * Query performace object Memory and locate Pool Paged Bytes and
   * Pool Nonpaged Bytes counters. The sum of these appears to be the
   * amount of memory consumed by the kernel.
   */
  for (int iterations = 0; /* true */; iterations++) {
    assert (iterations < PLT_MaxIterations);

    pDataBlk = PLTGetPerfData (acMemoryObj, &pPerfDataMem, &iPerfDataLen);
    pObjType = PLTFindPerfObject (pDataBlk, iMemoryObj);
    if (pObjType != NULL) break;
  } // get the performance object

  pPoolPagedBytesCtr    = PLTFindPerfCounter (pObjType, iPoolPagedBytesCtr);
  pPoolNonpagedBytesCtr = PLTFindPerfCounter (pObjType, iPoolNonpagedBytesCtr);

  pCntrBlk = PLTFirstCntrBlk (pObjType);
  (*pMemoryUsed)  = PLTGetPerfCounter (pCntrBlk, pPoolPagedBytesCtr);
  (*pMemoryUsed) += PLTGetPerfCounter (pCntrBlk, pPoolNonpagedBytesCtr);

  PLTReleasePerfMutex ();
}; // MEMGetKernelMemory (uint64)

/*******************************************************************\
| MEMGetTotalMemory                                                 |
\*******************************************************************/

void MEMGetTotalMemory (uint64 *pPhysicalUsed,
                        uint64 *pPhysicalFree,
                        uint64 *pSwapUsed)
{
  MEMORYSTATUS sSystemMemory;

  memset (&sSystemMemory, 0, sizeof (sSystemMemory));
  GlobalMemoryStatus (&sSystemMemory);

  (*pPhysicalFree) = sSystemMemory.dwAvailPhys;
  (*pPhysicalUsed) = sSystemMemory.dwTotalPhys - sSystemMemory.dwAvailPhys;
  (*pSwapUsed)     = sSystemMemory.dwTotalPageFile - sSystemMemory.dwAvailPageFile;
}; // MEMGetTotalMemory (uint64, uint64, uint64)

//-------------------------------------------------------------------
