/*******************************************************************\
| NET_Solaris_KStat.CC                                              |
+-------------------------------------------------------------------+
| The network information functions for Solaris.                    |
+-------------------------------------------------------------------+
| Copyright (C) Petr Tuma <petr.tuma@mff.cuni.cz>                   |
\*******************************************************************/

#include <kstat.h>

#include "com_defs.h"
#include "orb_defs.h"
#include "thr_defs.h"

#include "shr_defs.h"
#include "all_defs.h"

//-------------------------------------------------------------------

#define NET_MaxLOInterfaces     5
#define NET_MaxHMEInterfaces    10

uint64 aiLOSentPackets       [NET_MaxLOInterfaces];
uint64 aiLOReceivedPackets   [NET_MaxLOInterfaces];

uint64 aiHMESentBytes        [NET_MaxHMEInterfaces];
uint64 aiHMEReceivedBytes    [NET_MaxHMEInterfaces];
uint64 aiHMESentPackets      [NET_MaxHMEInterfaces];
uint64 aiHMEReceivedPackets  [NET_MaxHMEInterfaces];

//-------------------------------------------------------------------

/*******************************************************************\
| NETInit                                                           |
\*******************************************************************/

void NETInit (void)
{
  uint32 i;

  for (i = 0 ; i < NET_MaxLOInterfaces ; i ++)
  {
    aiLOSentPackets     [i] = 0;
    aiLOReceivedPackets [i] = 0;
  }

  for (i = 0 ; i < NET_MaxHMEInterfaces ; i ++)
  {
    aiHMESentBytes       [i] = 0;
    aiHMEReceivedBytes   [i] = 0;
    aiHMESentPackets     [i] = 0;
    aiHMEReceivedPackets [i] = 0;
  }
}

//-------------------------------------------------------------------

/*******************************************************************\
| NETGetTotalTraffic                                                |
\*******************************************************************/

void NETGetTotalTraffic (uint64 *pSentBytes, uint64 *pReceivedBytes,
                         uint64 *pSentPackets, uint64 *pReceivedPackets)
{
  kstat_ctl_t   *pKStats;
  kstat_t       *pKStat;
  kstat_named_t *pKValue;
  kid_t         iKStatID;
  int           iDummy;

  // We have to do this because kstat_lookup wants char pointer
  // instead of const char pointer which it probably should.

  char acLiteralSentBytes       [] = "obytes";
  char acLiteralReceivedBytes   [] = "rbytes";
  char acLiteralSentPackets     [] = "opackets";
  char acLiteralReceivedPackets [] = "ipackets";

  // First of all, open the kernel statistics control structure.

  pKStats = kstat_open ();
  assert (pKStats != NULL);
  iKStatID = pKStats->kc_chain_id;

  // Sum the packet counters for all interfaces in the kstat chain.

  while (TRUE)
  {
    (*pSentBytes)   = (*pReceivedBytes)   = 0;
    (*pSentPackets) = (*pReceivedPackets) = 0;

    for (pKStat = pKStats->kc_chain ;
         pKStat ;
         pKStat = pKStat->ks_next)
    {
      if (!strcmp (pKStat->ks_class, "net"))
      {
        // Because the kstat chain can change dynamically as we walk it,
        // watch the chain ID and restart the walk whenever the ID differs
        // from what we saw during the walk. The restart is done by breaking
        // from the cycle with pKStat not NULL.

        // The benchmarks run long enough to cause the counters to
        // overflow occasionally, which is why we attempt to
        // detect and compensate for that situation.

        #define UpdateValue(Iface,Data)                         \
        {                                                       \
          pKValue = (kstat_named_t *)                           \
                    kstat_data_lookup (                         \
                      pKStat, acLiteral##Data);                 \
          assert (pKValue != NULL);                             \
          assert (pKValue->data_type == KSTAT_DATA_UINT32);     \
          ai##Iface##Data [iKSInstance] +=                      \
            pKValue->value.ui32 -                               \
            (uint32) ai##Iface##Data [iKSInstance];              \
          (*p##Data) += ai##Iface##Data [iKSInstance];          \
        }

        if (!strcmp (pKStat->ks_module, "lo"))
        {
          // Interfaces lo have only packet counters

          uint32 iKSInstance = pKStat->ks_instance;
          assert (iKSInstance < NET_MaxLOInterfaces);

          if (kstat_read (pKStats, pKStat, NULL) != iKStatID) break;

          UpdateValue (LO, SentPackets);
          UpdateValue (LO, ReceivedPackets);
        }
        else if (!strcmp (pKStat->ks_module, "hme"))
        {
          // Interfaces hme have packet and byte counters

          uint32 iKSInstance = pKStat->ks_instance;
          assert (iKSInstance < NET_MaxHMEInterfaces);

          if (kstat_read (pKStats, pKStat, NULL) != iKStatID) break;

          UpdateValue (HME, SentPackets);
          UpdateValue (HME, ReceivedPackets);
          UpdateValue (HME, SentBytes);
          UpdateValue (HME, ReceivedBytes);
        }

        #undef UpdateValue
      }
    }

    // Leave the infinite cycle only if walk was not aborted prematurely
    // because of changes in the chain. Abort is indicated by pKStat not
    // reaching the end of the chain.

    if (pKStat)
    {
      iKStatID = kstat_chain_update (pKStats);
      assert (iKStatID > 0);
    }
    else break;
  };

  iDummy = kstat_close (pKStats);
  assert (!iDummy);
};

//-------------------------------------------------------------------

