/*******************************************************************\
| NET_Windows_PData.CC                                              |
+-------------------------------------------------------------------+
| The network information functions for Windows.                    |
+-------------------------------------------------------------------+
| Copyright (C) Lubomir Bulej <bulej@nenya.ms.mff.cuni.cz>          |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include <assert.h>

#include "com_defs.h"
#include "plt_defs.h"

//-------------------------------------------------------------------

#define NET_MaxIfaces             16

#define NET_OBJ_NetworkInterface  TEXT("Network Interface")
#define NET_CTR_BytesSent         TEXT("Bytes Sent/sec")
#define NET_CTR_PacketsSent       TEXT("Packets Sent/sec")
#define NET_CTR_BytesReceived     TEXT("Bytes Received/sec")
#define NET_CTR_PacketsReceived   TEXT("Packets Received/sec")

//-------------------------------------------------------------------

static DWORD iNetworkInterfaceObj;
static TCHAR acNetworkInterfaceObj [PLT_IndexDigits] = { TEXT ("") };

static DWORD iSentBytesCtr;
static DWORD iSentPacketsCtr;
static DWORD iReceivedBytessCtr;
static DWORD iReceivedPacketsCtr;

static uint64 aiSentBytes [NET_MaxIfaces];
static uint64 aiSentPackets [NET_MaxIfaces];
static uint64 aiReceivedBytes [NET_MaxIfaces];
static uint64 aiReceivedPackets [NET_MaxIfaces];

//-------------------------------------------------------------------

/*******************************************************************\
| NETInit                                                           |
\*******************************************************************/

void NETInit (void)
{
  PLTInitialize ();
  iNetworkInterfaceObj = PLTFindPerfIndex (NET_OBJ_NetworkInterface);
  _sntprintf (acNetworkInterfaceObj, PLT_IndexDigits, PLT_IndexFormat, iNetworkInterfaceObj);

  iSentBytesCtr       = PLTFindPerfIndex (NET_CTR_BytesSent);
  iSentPacketsCtr     = PLTFindPerfIndex (NET_CTR_PacketsSent);
  iReceivedBytesCtr   = PLTFindPerfIndex (NET_CTR_BytesReceived);
  iReceivedPacketsCtr = PLTFindPerfIndex (NET_CTR_PacketsReceived);

  memset (aiSentBytes, 0, sizeof (aiSentBytes));
  memset (aiSentPackets, 0, sizeof (aiSentPackets));
  memset (aiReceivedBytes, 0, sizeof (aiReceivedBytes));
  memset (aiReceivedPackets, 0, sizeof (aiReceivedPackets));
}; // NETInit (void)

//-------------------------------------------------------------------

/*******************************************************************\
| NETGetTotalTraffic                                                |
\*******************************************************************/

void NETGetTotalTraffic (uint64 *pSentBytes, uint64 *pReceivedBytes,
                         uint64 *pSentPackets, uint64 *pReceivedPackets)
{
  PERF_DATA_BLOCK *           pDataBlk;
  PERF_OBJECT_TYPE *          pObjType;
  PERF_COUNTER_BLOCK *        pCntrBlk;

  PERF_COUNTER_DEFINITION *   pSentBytesCtr;
  PERF_COUNTER_DEFINITION *   pSentPacketsCtr;
  PERF_COUNTER_DEFINITION *   pReceivedBytesCtr;
  PERF_COUNTER_DEFINITION *   pReceivedPacketsCtr;


  PLTAcquirePerfMutex ();

  /*
   * Query performace object Network Interface for packet and byte
   * counters. Sum counters from all instances.
   */
  for (int iterations = 0 ; /* true */ ; iterations++) {
    assert (iterations < PLT_MaxIterations);

    pDataBlk = PLTGetPerfData (acNetworkInterfaceObj, &pPerfDataMem, &iPerfDataLen);
    pObjType = PLTFindPerfObject (pDataBlk, iNetworkInterfaceObj);
    if (pObjType != NULL) break;
  }

  pSentBytesCtr       = PLTFindPerfCounter (pObjType, iSentBytesCtr);
  pSentPacketsCtr     = PLTFindPerfCounter (pObjType, iSentPacketsCtr);
  pReceivedBytesCtr   = PLTFindPerfCounter (pObjType, iReceivedBytesCtr);
  pReceivedPacketsCtr = PLTFindPerfCounter (pObjType, iReceivedPacketsCtr);

  (*pSentBytes)       = 0;
  (*pSentPackets)     = 0;
  (*pReceivedBytes)   = 0;
  (*pReceivedPackets) = 0;

  /*
   * Go through the instances and sum the counters. If there are no
   * instances of network interfaces, we return zeros... The following
   * code assumes that the order of interface instances does not change.
   */
  assert (pObjType->NumInstances >= 0);
  assert (pObjType->NumInstances <= NET_MaxIfaces);

  pCntrBlk = PLTFirstCntrBlk (pObjType);
  for (LONG iInstNum = 0; iInstNum < pObjType->NumInstances; iInstNum++) {
    /*
     * The benchmarks run long enough to cause the counters to
     * overflow occasionally, which is why we attempt to
     * detect and compensate for that situation.
     */
    #define UpdateValue(Cntr)                                     \
    {                                                             \
      ai##Cntr [iInstNum] +=                                      \
            (uint32) PLTGetPerfCounter (pCntrBlk, p##Cntr##Ctr) - \
            (uint32) ai##Cntr [iInstNum];                         \
      (*p##Cntr) += ai##Cntr [iInstNum];                          \
    }

    UpdateValue (SentBytes);
    UpdateValue (SentPackets);
    UpdateValue (ReceivedBytes);
    UpdateValue (ReceivedPackets);

    pCntrBlk = PLTNextCntrBlk (pCntrBlk);
    #undef UpdateValue
  };

  PLTReleasePerfMutex ();
}; // NETGetTotalTraffic (uint64, uint64, uint64, uint64)

//-------------------------------------------------------------------

