/*******************************************************************\
| PRO_Solaris_KStat.CC                                              |
+-------------------------------------------------------------------+
| The processor information functions for Solaris.                  |
+-------------------------------------------------------------------+
| Copyright (C) Petr Tuma <petr.tuma@mff.cuni.cz>                   |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include <dlfcn.h>
#include <kstat.h>
#include <thread.h>
#include <pthread.h>
#include <sys/sysinfo.h>
#include <sys/processor.h>

#include "com_defs.h"
#include "orb_defs.h"
#include "thr_defs.h"

#include "shr_defs.h"
#include "all_defs.h"

//-------------------------------------------------------------------

#define PRO_KStatCPU	"cpu_stat"

static THRLock          sPROMaximumThreadLock;

static thread_t         iPROMaximumSolarisThread = 0;
static pthread_t        iPROMaximumPosixThread   = 0;

static int (*pPROOriginalThrCreate) (void *, size_t, void *(*) (void *), void *, long, thread_t *) = NULL;
static int (*pPROOriginalPThreadCreate) (pthread_t *, const pthread_attr_t *, void *(*) (void *), void *) = NULL;

//-------------------------------------------------------------------

/*******************************************************************\
| PROInit                                                           |
\*******************************************************************/

void PROInit (void)
{
  THRLockCreate (&sPROMaximumThreadLock);

  THRLockLock (&sPROMaximumThreadLock);
  pPROOriginalThrCreate =
    (int (*) (void *, size_t, void *(*) (void *), void *, long, thread_t *))
    dlsym (RTLD_NEXT, "thr_create");
  pPROOriginalPThreadCreate =
    (int (*) (pthread_t *, const pthread_attr_t *, void *(*) (void *), void *))
    dlsym (RTLD_NEXT, "pthread_create");
  THRLockUnlock (&sPROMaximumThreadLock);
}

//-------------------------------------------------------------------

/*******************************************************************\
| PROGetApplicationThreads                                          |
\*******************************************************************/

void PROGetApplicationThreads (uint64 *pThreads)
{
  // There apparently is no way to query the number of threads
  // in the application. To work around this, the thread creation
  // functions are hooked to keep track of the highest thread ID that
  // was assigned. Here, thread IDs below the maximum are checked to
  // see whether the threads still exist. This might get terribly
  // inefficient if thread IDs are not reused ...

  // A race condition can occur if the initialization code of
  // the broker is called before the original pointers to the
  // hooked thread creation functions are saved. Hopefully that
  // will not happen ...

  // The algorithm can be improved by lowering the maximum ...

  sched_param sDummySchedParam;
  int         iDummyInt;

  (*pThreads) = 0;

  THRLockLock (&sPROMaximumThreadLock);

  for (thread_t iSolarisThread = 1 ;
       iSolarisThread < iPROMaximumSolarisThread ;
       iSolarisThread ++)
  {
    if (thr_getprio (iSolarisThread, &iDummyInt) == 0) (*pThreads) ++;
  };

  for (pthread_t iPosixThread = 1 ;
       iPosixThread < iPROMaximumPosixThread ;
       iPosixThread ++)
  {
    if (pthread_getschedparam (iPosixThread, &iDummyInt, &sDummySchedParam) == 0) (*pThreads) ++;
  }

  THRLockUnlock (&sPROMaximumThreadLock);
}

/*******************************************************************\
| thr_create                                                        |
\*******************************************************************/

extern "C" int thr_create (
  void *pStackBase, size_t iStackSize,
  void *(*pStart) (void *), void *pArgs,
  long iFlags, thread_t *pThread)
{
  THRLockLock (&sPROMaximumThreadLock);
  assert (pPROOriginalThrCreate != NULL);
  int iResult = (*pPROOriginalThrCreate) (pStackBase, iStackSize, pStart, pArgs, iFlags, pThread);
  if (iResult == 0)
  {
    if (pThread == NULL)
    {
      iPROMaximumSolarisThread ++;
    }
    else
    {
      if (iPROMaximumSolarisThread < (*pThread))
        iPROMaximumSolarisThread = (*pThread);
    }
  }
  THRLockUnlock (&sPROMaximumThreadLock);
  return (iResult);
}

/*******************************************************************\
| pthread_create                                                    |
\*******************************************************************/

extern "C" int pthread_create (
  pthread_t *pThread, const pthread_attr_t *pFlags,
  void *(*pStart) (void *), void *pArgs)
{
  THRLockLock (&sPROMaximumThreadLock);
  assert (pPROOriginalPThreadCreate != NULL);
  int iResult = (*pPROOriginalPThreadCreate) (pThread, pFlags, pStart, pArgs);
  if (iResult == 0)
  {
    if (iPROMaximumPosixThread < (*pThread))
      iPROMaximumPosixThread = (*pThread);
  };
  THRLockUnlock (&sPROMaximumThreadLock);
  return (iResult);
}

//-------------------------------------------------------------------

/*******************************************************************\
| PROGetTotalUsage                                                  |
\*******************************************************************/

void PROGetTotalUsage (uint64 *pKernelTicks,
                       uint64 *pWaitTicks,
                       uint64 *pUserTicks,
                       uint64 *pIdleTicks)
{
  kstat_ctl_t *pKStats;
  kstat_t     *pKStat;
  kid_t       iKStatID;
  int         iDummy;

  // First of all, open the kernel statistics control structure.

  pKStats = kstat_open ();
  assert (pKStats != NULL);
  iKStatID = pKStats->kc_chain_id;

  // Sum the tick counters for all processors in the kstat chain.
  // This can probably yield errorneous results now and then when
  // the tick counters overflow, but it should be rare enough case
  // to be ignored.

  while (TRUE)
  {
    (*pKernelTicks) = 0;
    (*pWaitTicks) = 0;
    (*pUserTicks) = 0;
    (*pIdleTicks) = 0;

    for (pKStat = pKStats->kc_chain ;
         pKStat ;
         pKStat = pKStat->ks_next)
    {
      if (!strncmp (pKStat->ks_name, PRO_KStatCPU, sizeof (PRO_KStatCPU) - 1))
      {
        // Because the kstat chain can change dynamically as we walk it,
        // watch the chain ID and restart the walk whenever the ID differs
        // from what we saw during the walk. The restart is done by breaking
        // from the cycle with pKStat not NULL.

        if (kstat_read (pKStats, pKStat, NULL) != iKStatID) break;

        cpu_stat_t &sKStatCPU = *((cpu_stat_t *) pKStat->ks_data);
        (*pKernelTicks) += sKStatCPU.cpu_sysinfo.cpu [CPU_KERNEL];
        (*pWaitTicks)   += sKStatCPU.cpu_sysinfo.cpu [CPU_WAIT];
        (*pUserTicks)   += sKStatCPU.cpu_sysinfo.cpu [CPU_USER];
        (*pIdleTicks)   += sKStatCPU.cpu_sysinfo.cpu [CPU_IDLE];
      }
    }

    // Leave the infinite cycle only if walk was not aborted prematurely
    // because of changes in the chain. Abort is indicated by pKStat not
    // reaching the end of the chain.

    if (pKStat)
    {
      iKStatID = kstat_chain_update (pKStats);
      assert (iKStatID > 0);
    }
    else break;
  }

  iDummy = kstat_close (pKStats);
  assert (!iDummy);
}

//-------------------------------------------------------------------
