/*******************************************************************\
| PRO_Windows_PData.CC                                              |
+-------------------------------------------------------------------+
| The processor information functions for Windows.                  |
+-------------------------------------------------------------------+
| Copyright (C) Lubomir Bulej <bulej@nenya.ms.mff.cuni.cz>          |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include <assert.h>

#include "com_defs.h"
#include "plt_defs.h"

//-------------------------------------------------------------------

#define PRO_OBJ_Process         TEXT("Process")
#define PRO_CTR_IDProcess       TEXT("ID Process")
#define PRO_CTR_ThreadCount     TEXT("Thread Count")

#define PRO_OBJ_Processor       TEXT("Processor")
#define PRO_CTR_UserTime        TEXT("% User Time")
#define PRO_CTR_ProcessorTime   TEXT("% Processor Time")
#define PRO_CTR_PrivilegedTime  TEXT("% Privileged Time")

//-------------------------------------------------------------------

static DWORD iProcessID;

static DWORD iProcessObj;
static TCHAR acProcessObj [PLT_IndexDigits]   = { TEXT ("") };

static DWORD iIDProcessCtr;
static DWORD iThreadCountCtr;

static DWORD iProcessorObj;
static TCHAR acProcessorObj [PLT_IndexDigits] = { TEXT ("") };

static DWORD iUserTimeCtr;
static DWORD iProcessorTimeCtr;
static DWORD iPrivilegedTimeCtr;

//-------------------------------------------------------------------

/*******************************************************************\
| PROInit                                                           |
\*******************************************************************/

void PROInit (void)
{
  PLTInitialize ();
  iProcessID = GetCurrentProcessId ();

  // Processor stats
  iProcessorObj = PLTFindPerfIndex (PRO_OBJ_Processor);
  _sntprintf (acProcessorObj, PLT_IndexDigits, PLT_IndexFormat, iProcessorObj);

  iUserTimeCtr       = PLTFindPerfIndex (PRO_CTR_UserTime);
  iProcessorTimeCtr  = PLTFindPerfIndex (PRO_CTR_ProcessorTime);
  iPrivilegedTimeCtr = PLTFindPerfIndex (PRO_CTR_PrivilegedTime);

  // Process processor stats
  iProcessObj = PLTFindPerfIndex (PRO_OBJ_Process);
  _sntprintf (acProcessObj, PLT_IndexDigits, PLT_IndexFormat, iProcessObj);

  iIDProcessCtr   = PLTFindPerfIndex (PRO_CTR_IDProcess);
  iThreadCountCtr = PLTFindPerfIndex (PRO_CTR_ThreadCount);
}; // PROInit (void)

//-------------------------------------------------------------------

/*******************************************************************\
| PROGetApplicationThreads                                          |
\*******************************************************************/

void PROGetApplicationThreads (uint64 *pThreads)
{
  PERF_DATA_BLOCK *           pDataBlk;
  PERF_OBJECT_TYPE *          pObjType;
  PERF_COUNTER_BLOCK *        pCntrBlk;

  PERF_COUNTER_DEFINITION *   pIDProcessCtr;
  PERF_COUNTER_DEFINITION *   pThreadCountCtr;


  PLTAcquirePerfMutex ();

  /*
   * Query Process performace object, search by process ID for instance for
   * this process and get the Thread Count counter.
   */
  for (int iterations = 0; /* true */; iterations++) {
    assert (iterations < PLT_MaxIterations);

    pDataBlk = PLTGetPerfData (acProcessObj, &pPerfDataMem, &iPerfDataLen);
    pObjType = PLTFindPerfObject (pDataBlk, iProcessObj);
    if (pObjType != NULL) break;
  } // get the performance object

  pIDProcessCtr   = PLTFindPerfCounter (pObjType, iIDProcessCtr);
  pThreadCountCtr = PLTFindPerfCounter (pObjType, iThreadCountCtr);

  /*
   * Find the current process instance and get its thread count.
   */
  pCntrBlk = PLTFirstCntrBlk (pObjType);
  for (LONG iInstNum = 0; iInstNum < pObjType->NumInstances; iInstNum++) {

    // look for process ID match
    if (iProcessID == PLTGetPerfCounter (pCntrBlk, pIDProcessCtr)) {
        (*pThreads) = PLTGetPerfCounter (pCntrBlk, pThreadCountCtr);

        PLTReleasePerfMutex ();
        return;
    };

    pCntrBlk = PLTNextCntrBlk (pCntrBlk);
  };

  assert (FALSE);
}; // PROGetApplicationThreads (uint64)

//-------------------------------------------------------------------

/*******************************************************************\
| PROGetTotalUsage                                                  |
\*******************************************************************/

void PROGetTotalUsage (uint64 *pKernelTicks,
                       uint64 *pWaitTicks,
                       uint64 *pUserTicks,
                       uint64 *pIdleTicks)
{
  PERF_DATA_BLOCK *           pDataBlk;
  PERF_OBJECT_TYPE *          pObjType;
  PERF_COUNTER_BLOCK *        pCntrBlk;

  PERF_COUNTER_DEFINITION *   pUserTimeCtr;
  PERF_COUNTER_DEFINITION *   pProcessorTimeCtr;
  PERF_COUNTER_DEFINITION *   pPrivilegedTimeCtr;


  PLTAcquirePerfMutex ();

  /*
   * Query performace object Processor and locate Processor, User and
   * Privileged Time counters. If several instances are found, add them
   * all up except instance named _Total.
   */
  for (int iterations = 0; /* true */; iterations++) {
    assert (iterations < PLT_MaxIterations);

    pDataBlk = PLTGetPerfData (acProcessorObj, &pPerfDataMem, &iPerfDataLen);
    pObjType = PLTFindPerfObject (pDataBlk, iProcessorObj);
    if (pObjType != NULL) break;
  } // get the performance object

  pUserTimeCtr       = PLTFindPerfCounter (pObjType, iUserTimeCtr);
  pProcessorTimeCtr  = PLTFindPerfCounter (pObjType, iProcessorTimeCtr);
  pPrivilegedTimeCtr = PLTFindPerfCounter (pObjType, iPrivilegedTimeCtr);

  (*pIdleTicks)   = 0;
  (*pUserTicks)   = 0;
  (*pWaitTicks)   = 0;
  (*pKernelTicks) = 0;

  /*
   * Sum up the data from all instances, including _Total.
   */
  assert (pObjType->NumInstances > 0);

  pCntrBlk = PLTFirstCntrBlk (pObjType);
  for (LONG iInstNum = 0; iInstNum < pObjType->NumInstances; iInstNum++) {
    (*pUserTicks)   += PLTGetPerfCounter (pCntrBlk, pUserTimeCtr);
    (*pIdleTicks)   += PLTGetPerfCounter (pCntrBlk, pProcessorTimeCtr);
    (*pKernelTicks) += PLTGetPerfCounter (pCntrBlk, pPrivilegedTimeCtr);

    pCntrBlk = PLTNextCntrBlk (pCntrBlk);
  };

  PLTReleasePerfMutex ();
}; // PROGetTotalUsage (uint64, uint64, uint64, uint64)

//-------------------------------------------------------------------
