/*******************************************************************\
| PLT_Windows_PData.CC                                              |
+-------------------------------------------------------------------+
| Platform-specific function to access Windows Performance Data.    |
+-------------------------------------------------------------------+
| Copyright (C) Lubomir Bulej <bulej@nenya.ms.mff.cuni.cz>          |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include <tchar.h>
#include <assert.h>
#include <malloc.h>
#include <stdlib.h>
#include <windows.h>
#include <winperf.h>
#include <winreg.h>

#include "plt_defs.h"

//-------------------------------------------------------------------

#define PLT_RKEY_Help         TEXT("HELP")
#define PLT_RKEY_Counter      TEXT("Counter")
#define PLT_RKEY_LastCounter  TEXT("Last Counter")
#define PLT_HKLM_Perflib      TEXT("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Perflib")
#define PLT_HKLM_PerflibEng   TEXT("SOFTWARE\\Microsoft\\Windows NT\\CurrentVersion\\Perflib\\009")

//-------------------------------------------------------------------

#define PLT_PerfMem_InitSize  128*1024
#define PLT_PerfMem_Increment 16*1024

//-------------------------------------------------------------------

DWORD iPerfDataLen = 0;
LPVOID pPerfDataMem = NULL;

static HANDLE hPerfMutex = NULL;
static DWORD iPerfMemPtrTLS = -1;
static DWORD iPerfMemLenTLS = -1;

//-------------------------------------------------------------------

static DWORD iLastCounter = 0;
static TCHAR * acNamesData = (TCHAR *) NULL;
static TCHAR ** apPerfNames = (TCHAR **) NULL;

//-------------------------------------------------------------------

/*******************************************************************\
| Private Functions                                                 |
\*******************************************************************/

VOID __GetPerfNames (VOID)
{
  HKEY    hPerflibKey;      // handle to the Perflib registry key
  DWORD   iMaxValueLen;     // size of the largest key value
  DWORD   iNameIndex;       // index of a performace object name

  DWORD   iDummyDword;
  LONG    iDummyLong;
  TCHAR * pCurrent;

  /*
   * First, get the number of counter items and allocate memory for
   * the array of pointers to performance object names.
   */
  iDummyLong = RegOpenKeyEx (HKEY_LOCAL_MACHINE, PLT_HKLM_Perflib, 0, KEY_READ, &hPerflibKey);
  assert (iDummyLong == ERROR_SUCCESS);

  iDummyDword = sizeof (iLastCounter);
  iDummyLong = RegQueryValueEx (hPerflibKey, PLT_RKEY_LastCounter, NULL, NULL, (BYTE *) &iLastCounter, &iDummyDword);
  assert (iDummyLong == ERROR_SUCCESS);

  iDummyLong = RegCloseKey (hPerflibKey);
  assert (iDummyLong == ERROR_SUCCESS);

  apPerfNames = (TCHAR **) calloc ((iLastCounter + 1), sizeof (TCHAR *));
  assert (apPerfNames != NULL);

  /*
   * Next, open the key containing performance object and counter names in
   * English and get memory for the all the strings based on the size of
   * the largest value (either "Counter" or "Help") in the key.
   */
  iDummyLong = RegOpenKeyEx (HKEY_LOCAL_MACHINE, PLT_HKLM_PerflibEng, 0, KEY_READ, &hPerflibKey);
  assert (iDummyLong == ERROR_SUCCESS);

  iDummyLong = RegQueryInfoKey (hPerflibKey, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &iMaxValueLen, NULL, NULL);
  assert (iDummyLong == ERROR_SUCCESS);

  iDummyDword = iMaxValueLen + 1;
  acNamesData = (TCHAR *) malloc (iDummyDword * sizeof (TCHAR));
  assert (acNamesData != NULL);

  /*
   * Last, read counter data and load performance object and
   * counter names into the names array by their index.
   */
  iDummyLong = RegQueryValueEx (hPerflibKey, PLT_RKEY_Counter, NULL, NULL, (BYTE *) acNamesData, &iDummyDword);
  assert (iDummyLong == ERROR_SUCCESS);

  iDummyLong = RegCloseKey (hPerflibKey);
  assert (iDummyLong == ERROR_SUCCESS);

  pCurrent = acNamesData;
  while (*pCurrent != (TCHAR) 0) {
    iNameIndex = _ttol (pCurrent);
    assert (iNameIndex <= iLastCounter);
    pCurrent += lstrlen (pCurrent) + 1;

    apPerfNames [iNameIndex] = pCurrent;
    pCurrent += lstrlen (pCurrent) + 1;
  };
}; // __GetPerfNames (void)

//-------------------------------------------------------------------

/*******************************************************************\
| Performance Data Navigation                                       |
\*******************************************************************/

/*
 * Performance data snapshot has the following layout:
 *
 * [PERF_DATA_BLOCK]
 *    [PERF_OBJECT_TYPE] (0 or more instances)
 *       [PERF_COUNTER_DEFINITION]
 *       [PERF_COUNTER_DEFINITION]
 *       ...
 *       [PERF_COUNTER_DEFINITION]
 *
 *       [PERF_INSTANCE_DEFINITION]
 *          [PERF_COUNTER_BLOCK]
 *             [raw counter data]
 *       ...
 *       [PERF_INSTANCE_DEFINITION]
 *          [PERF_COUNTER_BLOCK]
 *             [raw counter data]
 *    ...
 *    [PERF_OBJECT_TYPE] (-1 instances)
 *       [PERF_COUNTER_DEFINITION]
 *       [PERF_COUNTER_DEFINITION]
 *       ...
 *       [PERF_COUNTER_DEFINITION]
 *
 *       [PERF_COUNTER_BLOCK]
 *          [raw counter data]
 */

//-------------------------------------------------------------------

PERF_OBJECT_TYPE * PLTFirstObjType (PERF_DATA_BLOCK * pDataBlk)
{
  return ((PERF_OBJECT_TYPE *)
    ((BYTE *) pDataBlk + pDataBlk->HeaderLength));
}; // PLTFirstObjType (PERF_DATA_BLOCK *)


PERF_OBJECT_TYPE * PLTNextObjType (PERF_OBJECT_TYPE * pObjType)
{
  return ((PERF_OBJECT_TYPE *)
    ((BYTE *) pObjType + pObjType->TotalByteLength));
}; // PLTNextObjType (PERF_OBJECT_TYPE *)

//-------------------------------------------------------------------

PERF_COUNTER_DEFINITION * PLTFirstCntrDef (PERF_OBJECT_TYPE * pObjType)
{
  return ((PERF_COUNTER_DEFINITION *)
    ((BYTE *) pObjType + pObjType->HeaderLength));
}; // PLTFirstCntrDef (PERF_OBJECT_TYPE *)


PERF_COUNTER_DEFINITION * PLTNextCntrDef (PERF_COUNTER_DEFINITION * pCntrDef)
{
  return ((PERF_COUNTER_DEFINITION *)
    ((BYTE *) pCntrDef + pCntrDef->ByteLength));
}; // PLTNextCntrDef (PERF_COUNTER_DEFINITION *)

//-------------------------------------------------------------------

PERF_INSTANCE_DEFINITION * PLTFirstInstDef (PERF_OBJECT_TYPE * pObjType)
{
  return ((PERF_INSTANCE_DEFINITION *)
    ((BYTE *) pObjType + pObjType->DefinitionLength));
}; // PLTFirstInstDef (PERF_OBJECT_TYPE *)


PERF_INSTANCE_DEFINITION * PLTNextInstDef (PERF_INSTANCE_DEFINITION * pInstDef)
{
  PERF_COUNTER_BLOCK * pCntrBlk;
  pCntrBlk = PLTInstCntrBlk (pInstDef);

  return ((PERF_INSTANCE_DEFINITION *)
    ((BYTE *) pCntrBlk + pCntrBlk->ByteLength));
}; // PLTNextInstDef (PERF_INSTANCE_DEFINITION *)

//-------------------------------------------------------------------

PERF_COUNTER_BLOCK * PLTInstCntrBlk (PERF_INSTANCE_DEFINITION * pInstDef)
{
  return ((PERF_COUNTER_BLOCK *)
    ((BYTE *) pInstDef + pInstDef->ByteLength));
}; // PLTInstCntrBlk (PERF_INSTANCE_DEFINITION *)


PERF_COUNTER_BLOCK * PLTFirstCntrBlk (PERF_OBJECT_TYPE * pObjType)
{
  PERF_INSTANCE_DEFINITION * pInstDef;
  pInstDef = PLTFirstInstDef (pObjType);

  /*
   * The object never has any instances, its counter block is located
   * right after the object type definition. There is no instance
   * header.
   */
  if (pObjType->NumInstances == PERF_NO_INSTANCES)
    return (PERF_COUNTER_BLOCK *) pInstDef;

  /*
   * The object can have multiple instances, but they don't have to
   * be present. If there are any instances, we should return
   * counter block of the first instance.
   */
  if (pObjType->NumInstances > 0)
    return PLTInstCntrBlk (pInstDef);

  /*
   * Bail out if there is no instance, because there are no counters.
   */
  assert (FALSE);
  return NULL;
}; // PLTFirstCntrBlk (PERF_OBJECT_TYPE *)


/**
 * Returns counter block of an instance following this counter block.
 */
PERF_COUNTER_BLOCK * PLTNextCntrBlk (PERF_COUNTER_BLOCK * pCntrBlk)
{
  PERF_INSTANCE_DEFINITION * pInstDef;
  pInstDef = ((PERF_INSTANCE_DEFINITION *)
    ((BYTE *) pCntrBlk + pCntrBlk->ByteLength));

  return PLTInstCntrBlk (pInstDef);
}; // PLTNextCntrBlk (PERF_COUNTER_BLOCK *)

//-------------------------------------------------------------------

/*******************************************************************\
| Performance Data Lookup                                           |
\*******************************************************************/

DWORD PLTFindPerfIndex (const TCHAR * pName)
{
  assert (pName != NULL);
  assert (iLastCounter != 0);

  /*
   * Find the index of performance object or counter of given name.
   * Look only in even slots, because odd slots are used for help texts.
   */
  for (DWORD iPerfIndex = 0; iPerfIndex <= iLastCounter; iPerfIndex += 2) {
    if (apPerfNames [iPerfIndex] == NULL)
      continue;

    if (_tcsicmp (apPerfNames [iPerfIndex], pName) != 0)
      continue;

    return iPerfIndex;
  };

  assert (FALSE);
  return 0;
}; // PLTFindPerfIndex (const TCHAR *)


/**
 * Finds performance object in the performance data snapshot
 * using the object title index as the key. Returns NULL if
 * the specified object was not found.
 */
PERF_OBJECT_TYPE * PLTFindPerfObject (PERF_DATA_BLOCK * pDataBlk, DWORD iIndex)
{
  PERF_OBJECT_TYPE * pObjType;
  assert (pDataBlk != NULL);

  pObjType = PLTFirstObjType (pDataBlk);
  for (DWORD iObjNum = 0; iObjNum < pDataBlk->NumObjectTypes; iObjNum++) {
    if (pObjType->ObjectNameTitleIndex == iIndex)
        return pObjType;

    pObjType = PLTNextObjType (pObjType);
  };

  return NULL;
}; // PLTFindPerfObject (PERF_DATA_BLOCK *, DWORD)


/**
 * Finds counter object definition in object type data using
 * the counter object title index as the key.
 */
PERF_COUNTER_DEFINITION * PLTFindPerfCounter (PERF_OBJECT_TYPE * pObjType, DWORD iIndex)
{
  PERF_COUNTER_DEFINITION * pCntrDef;
  assert (pObjType != NULL);

  pCntrDef = PLTFirstCntrDef (pObjType);
  for (DWORD iCtrNum = 0; iCtrNum < pObjType->NumCounters; iCtrNum++) {
    if (pCntrDef->CounterNameTitleIndex == iIndex)
      return pCntrDef;

    pCntrDef = PLTNextCntrDef (pCntrDef);
  };

  assert (FALSE);
  return NULL;
}; // PLTFindPerfCounter (PERF_OBJECT_TYPE *, DWORD)

//-------------------------------------------------------------------

/*******************************************************************\
| Performance Data Retrieval                                        |
\*******************************************************************/

/**
 * Retrieves the specified data from registry. If the provided buffer
 * is too small, it is extended until the information fits in the
 * buffer. The pointer to the extended memory area along with its
 * size is written back to the caller.
 */
PERF_DATA_BLOCK * PLTGetPerfData (TCHAR * pValue, LPVOID * ppDataPtr, DWORD * piDataLen)
{
  BYTE    *pData = (BYTE *) (* ppDataPtr);
  DWORD   iSize = (* piDataLen);
  LONG    iResult;

  do {
    iResult = RegQueryValueEx (HKEY_PERFORMANCE_DATA, pValue, NULL, NULL, pData, &iSize);
    if (iResult == ERROR_MORE_DATA) {
      iSize = (* piDataLen) + PLT_PerfMem_Increment;
      pData = (BYTE *) realloc (pData, iSize);
      assert (pData != NULL);

      (* ppDataPtr) = (LPVOID) pData;
      (* piDataLen) = iSize;
      continue;
    }

    assert (iResult == ERROR_SUCCESS);
  } while (iResult != ERROR_SUCCESS);

  iResult = RegCloseKey (HKEY_PERFORMANCE_DATA);
  assert (iResult == ERROR_SUCCESS);

  return (PERF_DATA_BLOCK *) pData;
}; // PLTGetPerfData (TCHAR *, LPVOID *, DWORD *)


/**
 * Gets a performance counter from the data block. The size of the counter 
 * is determined from its definition. Only DWORD (32 bit) and LARGE (64 bit)
 * counter sizes are considered valid and returned.
 */
UINT64 PLTGetPerfCounter (PERF_COUNTER_BLOCK * pCntrBlk, PERF_COUNTER_DEFINITION * pCntrDef)
{
  DWORD   iCounterSize;
  BYTE    *pCounterData;

  pCounterData = (BYTE *) pCntrBlk + pCntrDef->CounterOffset;
  iCounterSize = pCntrDef->CounterType & PERF_SIZE_VARIABLE_LEN;
  if (iCounterSize == PERF_SIZE_DWORD)
      return *((DWORD *) pCounterData);

  if (iCounterSize == PERF_SIZE_LARGE)
      return *((UINT64 *) pCounterData);

  assert (FALSE);
  return 0;
}; // PLTGetPerfCounter (PERF_COUNTER_BLOCK *, PERF_COUNTER_DEFINITION *)

//-------------------------------------------------------------------

/*******************************************************************\
| Thread Local Memory Buffer                                        |
\*******************************************************************/

/**
 * Retrieves/initializes the thread-local size and pointer
 * to a memory buffer used for collecting performance data.
 */
VOID PLTGetPerfMemory (LPVOID * ppMemoryPtr, DWORD * piMemoryLen)
{
  DWORD   iMemoryLen;
  LPVOID  pMemoryPtr;

  pMemoryPtr = TlsGetValue (iPerfMemPtrTLS);
  assert (GetLastError () == NO_ERROR);

  iMemoryLen = (DWORD) TlsGetValue (iPerfMemLenTLS);
  assert (GetLastError () == NO_ERROR);

  if (pMemoryPtr == NULL && iMemoryLen == 0) {
    iMemoryLen = PLT_PerfMem_InitSize;
    pMemoryPtr = malloc (iMemoryLen);
    assert (pMemoryPtr != NULL);

    PLTPutPerfMemory (pMemoryPtr, iMemoryLen);
  }

  (* ppMemoryPtr) = pMemoryPtr;
  (* piMemoryLen) = iMemoryLen;
} // PLTGetPerfMemory (LPVOID *, DWORD *)


/**
 * Stores the thread-local size and pointer to a memory buffer.
 */
VOID PLTPutPerfMemory (LPVOID pMemoryPtr, DWORD iMemoryLen)
{
  BOOL bResult;

  bResult = TlsSetValue (iPerfMemPtrTLS, pMemoryPtr);
  assert (bResult);

  bResult = TlsSetValue (iPerfMemLenTLS, (LPVOID) iMemoryLen);
  assert (bResult);
} // PLTPutPerfMemory (LPVOID, DWORD)


/**
 * Releases the thread-local memory buffer.
 */
VOID PLTDelPerfMemory (VOID)
{
  DWORD   iMemoryLen;
  LPVOID  pMemoryPtr;

  PLTGetPerfMemory (&pMemoryPtr, &iMemoryLen);
  if (pMemoryPtr != NULL && iMemoryLen != 0) {
    free (pMemoryPtr);
  }

  PLTPutPerfMemory (NULL, 0);
} // PLTDelPerfMemory (VOID)

//-------------------------------------------------------------------

/*******************************************************************\
| Performance Data Mutex                                            |
\*******************************************************************/

VOID PLTAcquirePerfMutex (VOID)
{
  DWORD iResult = WaitForSingleObject (hPerfMutex, INFINITE);
  assert (iResult == WAIT_OBJECT_0);
} // PLTGetPerfMutex (VOID)


VOID PLTReleasePerfMutex (VOID)
{
  BOOL bResult = ReleaseMutex (hPerfMutex);
  assert (bResult);
} // PLTPutPerfMutex (VOID)

//-------------------------------------------------------------------

/*******************************************************************\
| Platform Library Initialization                                   |
\*******************************************************************/

/**
 * Initializes the platform library. Creates the mutex,
 * allocates performance data buffer and TLS indices and
 * retrieves the names of performance objects.
 */
VOID PLTInitialize (VOID)
{
  hPerfMutex = CreateMutex (NULL, FALSE, TEXT ("Xampler_PerfMutex"));
  assert (hPerfMutex != NULL);
  PLTAcquirePerfMutex ();

  // alloc buffer for performance data
  if (pPerfDataMem == NULL && iPerfDataLen == 0) {
    iPerfDataLen = PLT_PerfMem_InitSize;
    pPerfDataMem = malloc (iPerfDataLen);
    assert (pPerfDataMem != NULL);
  }

  // alloc thread local storage
  if (iPerfMemPtrTLS == -1) {
    iPerfMemPtrTLS = TlsAlloc ();
    assert (iPerfMemPtrTLS != -1);
  }

  if (iPerfMemLenTLS == -1) {
    iPerfMemLenTLS = TlsAlloc ();
    assert (iPerfMemLenTLS != -1);
  }

  // get performance object names
  if (iLastCounter == 0) {
    __GetPerfNames ();
    assert (iLastCounter != 0);
  }

  PLTReleasePerfMutex ();
} // PLTInitialize (VOID)

//-------------------------------------------------------------------
