/*******************************************************************\
| TIM_Linux_HPT.CC                                                  |
+-------------------------------------------------------------------+
| The timer functions for Linux platforms with RDTSC or ITC.        |
+-------------------------------------------------------------------+
| Copyright (C) Petr Tuma <petr.tuma@mff.cuni.cz>                   |
|               Adam Buble <adam.buble@mff.cuni.cz>                 |
|               Lubomir Bulej <lubomir.bulej@mff.cuni.cz>           |
\*******************************************************************/

#include <unistd.h>
#include <sys/time.h>

#include "com_defs.h"
#include "orb_defs.h"
#include "thr_defs.h"

#include "shr_defs.h"
#include "all_defs.h"

//-------------------------------------------------------------------

#define TIM_MaxLineLen		1024

// TIM_CycleScale         How many samples to get scale
// TIM_CycleGranularity	  How many samples to get granularity
// TIM_MinimumPrecision	  Inverse value of resolution precision

#define TIM_CycleScale        15
#define TIM_CycleGranularity  100
#define TIM_MinimumPrecision  1000

// TIM_SleepInit          How long sleep to callibrate on initially
// TIM_SleepExtend        How many failures warrant extending sleep

#define TIM_SleepInit         1
#define TIM_SleepExtend       10

//-------------------------------------------------------------------

uint64 iTIMTimerScale       = 0;
uint64 iTIMTimerGranularity = 0;

//-------------------------------------------------------------------

/*******************************************************************\
| TIMInit                                                           |
\*******************************************************************/

void TIMInit (void)
{
  struct timeval        sStartTime, sEndTime;
  uint64                iStartTime, iEndTime;
  uint64                iCountDelta, iClockDelta;
  uint64                iSumScale, iAvgScale, iNewScale, iCPUScale;
  uint64                iDifference;
  int                   iSumCount;
  int                   iErrCount;

  FILE                  *pFile;
  char                  acLine [TIM_MaxLineLen];
  double                fCPUScale;
  int                   iDummy;

  #if defined __i386__

    // First of all we have to test if the processor supports the RDTSC
    // instruction. This is done in three steps, first the existence of the
    // CPUID instruction is checked by trying to flip bit 21 of EFLAGS,
    // second the ability of the CPUID instruction to provide feature list
    // is checked, third the feature list is used to check for existence of
    // RDTSC.

    uint32                iEFlagsOne, iEFlagsTwo;
    uint32                iEAX, iEDX;
    int                   iHasRDTSC = FALSE;

    asm volatile ("pushfl ; popl %0" : "=r,m" (iEFlagsOne));
    iEFlagsOne ^= (1 << 21);
    asm volatile ("pushl %0 ; popfl" : : "r,m" (iEFlagsOne));
    asm volatile ("pushfl ; popl %0" : "=r,m" (iEFlagsTwo));

    if (!((iEFlagsOne ^ iEFlagsTwo) & (1 << 21)))
    {
      iEAX = 0;
      asm volatile ("cpuid" : "+a" (iEAX) : : "ebx", "ecx", "edx");

      if (iEAX >= 1)
      {
        iEAX = 1;
        asm volatile ("cpuid" : "+a" (iEAX), "=d" (iEDX): : "ebx", "ecx");

        if (iEDX & (1 << 4))
        {
          iHasRDTSC = TRUE;
        }
      }
    }

    // Everyone should support RDTSC these days.

    assert (iHasRDTSC);

  #endif

  // Check whether the processor frequency in the proc filesystem is
  // available. Note that this can change when power management is
  // enabled, hopefully everybody turns power management off !

  fCPUScale = 0;

  pFile = fopen ("/proc/cpuinfo", "r");
  assert (pFile != NULL);
  while (fgets (acLine, sizeof (acLine), pFile) != NULL)
  {
    #if   defined __i386__
      sscanf (acLine, "cpu MHz : %lf", &fCPUScale);
    #elif defined __ia64__
      sscanf (acLine, "itc MHz : %lf", &fCPUScale);
    #else
      #error
    #endif
  }
  iDummy = fclose (pFile);
  assert (!iDummy);

  iCPUScale = (uint64) (fCPUScale * 1E6);

  if (iCPUScale)
  {
    iTIMTimerScale = iCPUScale;
  }
  else
  {
    // We seem to lack the processor frequency in the proc file system.
    // Time a short interfal using the official system timer, and
    // calculate the resolution of the RDTSC timer from that. Do
    // it several times to avoid using an incorrect result due
    // to random distortions.

    // This algorithm does not really guarantee calculating the
    // correct divisor, but then again, if scheduler interrupts
    // can happen at any time, no algorithm really can.

    iSumScale = 0;
    iAvgScale = 0;
    iSumCount = 0;
    iErrCount = 0;

    while (iSumCount < TIM_CycleScale)
    {
      iStartTime = TIMGetTimestamp ();
      gettimeofday (&sStartTime, NULL);
      THRThreadSleep (TIM_SleepInit + iErrCount / TIM_SleepExtend);
      iEndTime = TIMGetTimestamp ();
      gettimeofday (&sEndTime, NULL);

      iCountDelta = (iEndTime - iStartTime) * 1000000uLL;
      iClockDelta = ((uint64) sEndTime.tv_sec * 1000000uLL + (uint64) sEndTime.tv_usec) -
                    ((uint64) sStartTime.tv_sec * 1000000uLL + (uint64) sStartTime.tv_usec);

      iNewScale   = iCountDelta / iClockDelta;
      iSumScale   += iNewScale;
      iSumCount   ++;
      iAvgScale   = iSumScale / iSumCount;

      if (iNewScale < iAvgScale) iDifference = iAvgScale - iNewScale;
                            else iDifference = iNewScale - iAvgScale;
      if (iDifference && ((iAvgScale / iDifference) < TIM_MinimumPrecision))
      {
        iSumScale = 0;
        iSumCount = 0;
        iErrCount ++;
      }
    }

    iTIMTimerScale = iAvgScale;
  }

  // Granularity is always 1 with RDTSC.

  iTIMTimerGranularity = 1;
}

/*******************************************************************\
| TIMGetTimestamp                                                   |
\*******************************************************************/

uint64 TIMGetTimestamp (void)
{
  uint64 iResult;

  #if   defined __i386__
    asm volatile ("rdtsc" : "=A" (iResult));
  #elif defined __ia64__
    asm volatile ("mov %0=ar.itc" : "=r"(iResult) :: "memory");
  #else
    #error
  #endif

  return (iResult);
}

//-------------------------------------------------------------------

