/*******************************************************************\
| Client.CC (Dispatch Servant Instances)                            |
+-------------------------------------------------------------------+
| Copyright (C) Petr Tuma <petr.tuma@mff.cuni.cz>                   |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include "com_defs.h"
#include "orb_defs.h"
#include "thr_defs.h"

#include "shr_defs.h"
#include "all_defs.h"

#include "defs.h"

//-------------------------------------------------------------------

#define BEN_MaxLineLen          1024

#define BEN_MinInstances        1
#define BEN_MaxInstances        SAM_MaxInstances
#define BEN_MinThreads          1
#define BEN_MaxThreads          SAM_MaxThreads
#define BEN_StepLimit           5
#define BEN_StepScale           10

// Do not use CORBA var classes as global variables, it causes
// problems with proper initialization and shutdown order.

tPingSequence   *pPingSequence;
CORBA::ULong    iPingSequenceLength;

//-------------------------------------------------------------------

/*******************************************************************\
| BENGetName                                                        |
\*******************************************************************/

const char *BENGetName (void)
{
  return ("Dispatch Servant Instances Client");
};

//-------------------------------------------------------------------

/*******************************************************************\
| ActionPing                                                        |
\*******************************************************************/

void ActionPing (tMEAArgs *pArgs ORBDclNext) ORBDclThrow
{
  (*pPingSequence) [RNDGetRandom (iPingSequenceLength)]->Pong (ORBArgOnly);
};

//-------------------------------------------------------------------

/*******************************************************************\
| BENClient                                                         |
\*******************************************************************/

void BENClient (
  CORBA::Object_ptr pReference,
  int iArgC, char *apArgV [])
{
  PingFactory_ptr       pPingFactory;
  tPingSequence_var     vPingSequenceAddition;

  uint32                iInstancesMin = BEN_MinInstances, iInstancesMax = BEN_MaxInstances;
  uint32                iInstancesLimit = BEN_StepLimit, iInstancesScale = BEN_StepScale;
  uint32                iInstancesStep, iInstancesCount;
  uint32                iInstance;

  uint32                iThreadsMin = BEN_MinThreads, iThreadsMax = BEN_MaxThreads;
  uint32                iThreadsLimit = BEN_StepLimit, iThreadsScale = BEN_StepScale;
  uint32                iThreadsStep, iThreadsCount;

  char                  acLine [BEN_MaxLineLen];

  // Parse the command line

  switch (iArgC)
  {
    case 9:
      iThreadsMin       = atol (apArgV [5]);
      iThreadsMax       = atol (apArgV [6]);
      iThreadsLimit     = atol (apArgV [7]);
      iThreadsScale     = atol (apArgV [8]);
    case 5:
      iInstancesMin     = atol (apArgV [1]);
      iInstancesMax     = atol (apArgV [2]);
      iInstancesLimit   = atol (apArgV [3]);
      iInstancesScale   = atol (apArgV [4]);
    case 1:
      break;
    default:
      DSPFatal ("BEN", "Usage: " << apArgV [0] << " [InstancesMin Max Limit Scale [ThreadsMin Max Limit Scale]]");
  };

  // Execute the measurement

  ORBEnvVar
  ORBExcTry
  {
    pPingFactory = PingFactory::_narrow (pReference);

    // Originally, this was implemented as calling all objects from
    // the sequence. For a large number of objects, however, this
    // took too long time to complete.

    // An alternative, applied here, is to call a single random object
    // as an action. Provided that the scaling mechanism will call the
    // action enough times, the results should be similar.

    pPingSequence = new tPingSequence ();
    iPingSequenceLength = 0;

    MEAInitStep (iInstancesStep, iInstancesCount, iInstancesMin, iInstancesScale);
    while (iInstancesCount <= iInstancesMax)
    {
      // Get additional objects for this step of the benchmark and append
      // them at the end of the sequence holding all objects. Be sure to
      // delete the sequence holding the additional objects at once so that
      // it does not distort the measurement.

      vPingSequenceAddition = pPingFactory->Create (iInstancesCount - iPingSequenceLength ORBArgNext); ORBExcCheck
      pPingSequence->length (iInstancesCount);
      for (iInstance = 0 ;
           iInstance < vPingSequenceAddition->length () ;
           iInstance ++)
      {
        (*pPingSequence) [iPingSequenceLength + iInstance] =
          Ping::_duplicate (vPingSequenceAddition [(CORBA::ULong) iInstance]);
      };
      iPingSequenceLength = iInstancesCount;
      vPingSequenceAddition = NULL;

      // Execute the measurement for all thread numbers.

      MEAInitStep (iThreadsStep, iThreadsCount, iThreadsMin, iThreadsScale);
      while (iThreadsCount <= iThreadsMax)
      {
        snprintf (acLine, sizeof (acLine),
                  "[1]\tInstances\t%lu\t"
                  "[1]\tThreads\t%lu\t",
                  iInstancesCount,
                  iThreadsCount);
        MEAMeasure (acLine,
                    ActionPing,
                    NULL, 0,
                    iThreadsCount,
                    SAM_TestSamples
                    ORBArgNext); ORBExcCheck

        MEANextStep (iThreadsStep, iThreadsCount, iThreadsLimit, iThreadsScale);
      };

      // There is no deallocation of servants.

      MEANextStep (iInstancesStep, iInstancesCount, iInstancesLimit, iInstancesScale);
    };
  }
  ORBExcCatch
  {
    DSPFatal ("BEN", "Client " << ORBExcShow)
  };
};

//-------------------------------------------------------------------
