/*******************************************************************\
| Client.CC (Marshal Octet Sequence)                                |
+-------------------------------------------------------------------+
| Copyright (C) Petr Tuma <petr.tuma@mff.cuni.cz>                   |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

#include "com_defs.h"
#include "orb_defs.h"
#include "thr_defs.h"

#include "shr_defs.h"
#include "all_defs.h"

#include "defs.h"

//-------------------------------------------------------------------

#define BEN_MaxLineLen          1024

#define BEN_MinOctets           1
#define BEN_MaxOctets           SAM_MaxOctets
#define BEN_MinThreads          1
#define BEN_MaxThreads          SAM_MaxThreads
#define BEN_StepLimit           5
#define BEN_StepScale           10

// Do not use CORBA var classes as global variables, it causes
// problems with proper initialization and shutdown order.

Ping_ptr        pPing;

#if   defined BENDirectionIN
  tOctetSequence *pPingOctetSequence;
#elif defined BENDirectionOUT
  tOctetSequence *pPingOctetSequence;
  CORBA::ULong   iPingOctetSequenceLength;
#elif defined BENDirectionINOUT
  tOctetSequence *pPingOctetSequence;
#elif defined BENDirectionRETURN
  CORBA::ULong   iPingOctetSequenceLength;
#else
  #error
#endif

//-------------------------------------------------------------------

/*******************************************************************\
| BENGetName                                                        |
\*******************************************************************/

const char *BENGetName (void)
{
  return ("Marshal Octet Sequence Client");
};

//-------------------------------------------------------------------

/*******************************************************************\
| ActionPing                                                        |
\*******************************************************************/

void ActionPing (tMEAArgs *pArgs ORBDclNext) ORBDclThrow
{
  #if   defined BENDirectionIN
    pPing->Pong (*pPingOctetSequence ORBArgNext);
  #elif defined BENDirectionOUT
    pPing->Pong (iPingOctetSequenceLength, pPingOctetSequence ORBArgNext);
    delete (pPingOctetSequence);
  #elif defined BENDirectionINOUT
    pPing->Pong (*pPingOctetSequence ORBArgNext);
  #elif defined BENDirectionRETURN
    delete (pPing->Pong (iPingOctetSequenceLength ORBArgNext));
  #else
    #error
  #endif
};

//-------------------------------------------------------------------

/*******************************************************************\
| BENClient                                                         |
\*******************************************************************/

void BENClient (
  CORBA::Object_ptr pReference,
  int iArgC, char *apArgV [])
{
  uint32                iOctetsMin = BEN_MinOctets, iOctetsMax = BEN_MaxOctets;
  uint32                iOctetsLimit = BEN_StepLimit, iOctetsScale = BEN_StepScale;
  uint32                iOctetsStep, iOctetsCount;

  uint32                iThreadsMin = BEN_MinThreads, iThreadsMax = BEN_MaxThreads;
  uint32                iThreadsLimit = BEN_StepLimit, iThreadsScale = BEN_StepScale;
  uint32                iThreadsStep, iThreadsCount;

  char                  acLine [BEN_MaxLineLen];

  // Parse the command line

  switch (iArgC)
  {
    case 9:
      iThreadsMin       = atol (apArgV [5]);
      iThreadsMax       = atol (apArgV [6]);
      iThreadsLimit     = atol (apArgV [7]);
      iThreadsScale     = atol (apArgV [8]);
    case 5:
      iOctetsMin     = atol (apArgV [1]);
      iOctetsMax     = atol (apArgV [2]);
      iOctetsLimit   = atol (apArgV [3]);
      iOctetsScale   = atol (apArgV [4]);
    case 1:
      break;
    default:
      DSPFatal ("BEN", "Usage: " << apArgV [0] << " [OctetsMin Max Limit Scale [ThreadsMin Max Limit Scale]]");
  };

  // Execute the measurement

  ORBEnvVar
  ORBExcTry
  {
    pPing = Ping::_narrow (pReference);

    MEAInitStep (iOctetsStep, iOctetsCount, iOctetsMin, iOctetsScale);
    while (iOctetsCount <= iOctetsMax)
    {
      // Allocate the sequence to be passed if necessary.

      #if   defined BENDirectionIN
        pPingOctetSequence = new tOctetSequence (iOctetsCount);
        pPingOctetSequence->length (iOctetsCount);
      #elif defined BENDirectionOUT
        iPingOctetSequenceLength = iOctetsCount;
      #elif defined BENDirectionINOUT
        pPingOctetSequence = new tOctetSequence (iOctetsCount);
        pPingOctetSequence->length (iOctetsCount);
      #elif defined BENDirectionRETURN
        iPingOctetSequenceLength = iOctetsCount;
      #else
        #error
      #endif

      // Execute the measurement for all thread numbers.

      MEAInitStep (iThreadsStep, iThreadsCount, iThreadsMin, iThreadsScale);
      while (iThreadsCount <= iThreadsMax)
      {
        snprintf (acLine, sizeof (acLine),
                  "[STR]\tDirection\t" BENDirectionQuoted "\t"
                  "[1]\tOctets\t%lu\t"
                  "[1]\tThreads\t%lu\t",
                  iOctetsCount,
                  iThreadsCount);
        MEAMeasure (acLine,
                    ActionPing,
                    NULL, 0,
                    iThreadsCount,
                    SAM_TestSamples
                    ORBArgNext); ORBExcCheck

        MEANextStep (iThreadsStep, iThreadsCount, iThreadsLimit, iThreadsScale);
      };

      // Free the sequence to be passed if necessary.

      #if   defined BENDirectionIN
        delete (pPingOctetSequence);
      #elif defined BENDirectionOUT
      #elif defined BENDirectionINOUT
        delete (pPingOctetSequence);
      #elif defined BENDirectionRETURN
      #else
        #error
      #endif

      MEANextStep (iOctetsStep, iOctetsCount, iOctetsLimit, iOctetsScale);
    };
  }
  ORBExcCatch
  {
    DSPFatal ("BEN", "Client " << ORBExcShow)
  };
};

//-------------------------------------------------------------------

