/*******************************************************************\
| SHR_Defs.H                                                        |
+-------------------------------------------------------------------+
| Definitions for all shared modules.                               |
+-------------------------------------------------------------------+
| Copyright (C) Petr Tuma <petr.tuma@mff.cuni.cz>                   |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

//-------------------------------------------------------------------

/*******************************************************************\
| Basics                                                            |
\*******************************************************************/

#define FALSE             0
#define TRUE              1

#define sizeofarray(aArr)       (sizeof (aArr) / sizeof (aArr [0]))

//-------------------------------------------------------------------

/*******************************************************************\
| Configuration                                                     |
\*******************************************************************/

// Macros to handle different versions of the stream library

#if   defined DSPStreamsOLD
  #define DSPIStream    istream
  #define DSPOStream    ostream
  #define DSPIOStream   iostream
  #define DSPEndLine    endl
  #define DSPFlush      flush
  #define DSPStdOut     cout
  #define DSPStdErr     cerr
#elif defined DSPStreamsNEW
  #define DSPIStream    std::istream
  #define DSPOStream    std::ostream
  #define DSPIOStream   std::iostream
  #define DSPEndLine    std::endl
  #define DSPFlush      std::flush
  #define DSPStdOut     std::cout
  #define DSPStdErr     std::cerr
#else
  #error
#endif

// Macros to handle different approaches to exception handling

#if   defined ORBExceptionsCPP
  #define ORBSubIfExc(X)        X
  #define ORBSubIfEnv(X)
  #define ORBComIfExc           ,
  #define ORBComIfEnv
#elif defined ORBExceptionsORB
  #define ORBSubIfExc(X)
  #define ORBSubIfEnv(X)        X
  #define ORBComIfExc
  #define ORBComIfEnv           ,
#else
  #error
#endif

#define ORBEnvVar               ORBSubIfEnv (CORBA::Environment oEnv;)
#define ORBEnvArg               ORBSubIfEnv (oEnv)

#define ORBExcTry               ORBSubIfExc (try)
#define ORBExcCheck             ORBSubIfEnv (if (oEnv.exception ()) goto __ORBExcCatch;)
#define ORBExcCatch             ORBSubIfExc (catch (const CORBA::SystemException &sEx)) \
                                ORBSubIfEnv (__ORBExcCatch: if (CORBA::Exception *pEx = oEnv.exception ()))
#define ORBExcPass              ORBSubIfExc (catch (...) { throw; };) \
                                ORBSubIfEnv (__ORBExcCatch: ;)
#define ORBExcShow              ORBSubIfExc (ORBExcStream (sEx)) \
                                ORBSubIfEnv (ORBExcStream (*CORBA::SystemException::_downcast (pEx)))

#define ORBDclOnly              ORBSubIfExc (void) \
                                ORBSubIfEnv (CORBA::Environment &oEnv)
#define ORBDclNext              ORBComIfEnv ORBSubIfEnv (CORBA::Environment &oEnv)
#define ORBDclThrow             ORBSubIfExc (throw (CORBA::SystemException))

#define ORBArgOnly              ORBSubIfEnv (oEnv)
#define ORBArgNext              ORBComIfEnv ORBSubIfEnv (oEnv)

//-------------------------------------------------------------------

/*******************************************************************\
| Test Suite Modules                                                |
\*******************************************************************/

void BENClient (CORBA::Object_ptr pReference, int iArgC, char *apArgV []);
CORBA::Object_ptr BENServer (int iArgC, char *apArgV []);
const char *BENGetName (void);

//-------------------------------------------------------------------

/*******************************************************************\
| Platform Independent Modules                                      |
\*******************************************************************/

/*******************************************************************\
| Display Definitions                                               |
\*******************************************************************/

// DSP_HeaderBenchie    Header for benchmark output
// DSP_HeaderConfig     Header for configuration output
// DSP_HeaderMeasure    Header for measurement output
// DSP_HeaderSample     Header for sample output

#define DSP_HeaderBenchie       "[BENCHMARK]\t"
#define DSP_HeaderConfig        "[CONFIGURATION]\t"
#define DSP_HeaderMeasure       "[MEASUREMENT]\t"
#define DSP_HeaderSample        "[SAMPLE]\t"

// DSPFatal		Displays a fatal error message and exits

#define DSPFatal(Cls,Msg)                                           \
{                                                                   \
  DSPStdErr << "[" Cls "] " << Msg << DSPEndLine;                   \
  exit (1);                                                         \
}

/*******************************************************************\
| Measurement Definitions                                           |
\*******************************************************************/

// pMEAServer           Server object

extern Server_ptr pMEAServer;

// tMEAStamp            Description of stamp

#define MEA_StampLocal          1
#define MEA_StampRemote         2
#define MEA_StampAbsolute       4

struct tMEAStamp
{
  char *pName;
  int  iFlags;
};

// tMEAArgs             Arguments used by action
// MEAAction            Measurement action

#define MEA_MaxStamps           10

struct tMEAArgs
{
  uint64 aiTimes [MEA_MaxStamps];
};

typedef void MEAAction (tMEAArgs * ORBDclNext);

// MEAUsage             Execute usage measurement
// MEAMeasure		Execute arbitrary measurement

void MEAUsage   (void);
void MEAMeasure (char *pName,
                 MEAAction *pAction,
                 tMEAStamp *asStamps, uint32 iStamps,
                 uint32 iThreads,
                 uint32 iSamples
                 ORBDclNext) ORBDclThrow;

// MEAInitStep          Initializes a sequence of measurements
// MEANextStep          Advances a sequence of measurements

#define MEAInitStep(Step,Current,Min,Scale)     \
{                                               \
  Current = Min;                                \
  Step = 1;                                     \
  while ((Step) < (Min)) (Step) *= (Scale);	\
}

#define MEANextStep(Step,Current,Limit,Scale)	\
{                                               \
  if (Step * Limit > Current)                   \
  {                                             \
    Current += Step;                            \
  }                                             \
  else                                          \
  {                                             \
    Step *= Scale;                              \
    Current = Step;                             \
  };                                            \
}

// This template is useful for working with items whose
// minimum, average and maximum values are taken.

template <typename tValue> class tMEAMinAvgMax
{
  public:
    tValue      iMin;
    tValue      iAvg;
    tValue      iMax;
    int         iInitialized;

    /*******************************************************************\
    | tMEAMinAvgMax::Initialize                                         |
    \*******************************************************************/

    void Initialize (void)
    {
      iInitialized = FALSE;
      iMin = 0;
      iAvg = 0;
      iMax = 0;
    };

    /*******************************************************************\
    | tMEAMinAvgMax::Update (tValue)                                    |
    \*******************************************************************/

    void Update (tValue iValue)
    {
      if (!iInitialized)
      {
        iAvg = iValue;
        iMin = iValue;
        iMax = iValue;
        iInitialized = TRUE;
      }
      else
      {
        iAvg += iValue;
        if (iValue < iMin) iMin = iValue;
        if (iValue > iMax) iMax = iValue;
      };
    };

    /*******************************************************************\
    | tMEAMinAvgMax::Update (tMEAMinAvgMax)                             |
    \*******************************************************************/

    void Update (tMEAMinAvgMax<tValue> &oValue)
    {
      if (oValue.iInitialized)
      {
        if (!iInitialized)
        {
          iAvg = oValue.iAvg;
          iMin = oValue.iMin;
          iMax = oValue.iMax;
          iInitialized = TRUE;
        }
        else
        {
          iAvg += oValue.iAvg;
          if (oValue.iMin < iMin) iMin = oValue.iMin;
          if (oValue.iMax > iMax) iMax = oValue.iMax;
        };
      };
    };

    /*******************************************************************\
    | tMEAMinAvgMax::Normalize                                          |
    \*******************************************************************/

    void Normalize (uint64 iCount)
    {
      iAvg /= iCount;
    };

    /*******************************************************************\
    | tMEAMinAvgMax::Print                                              |
    \*******************************************************************/

    void Print (const char *pHeader)
    {
      DSPStdOut << "[3]\t" << pHeader << '\t'
                << iMin << '\t'
                << iAvg << '\t'
                << iMax << '\t';
    };
};

//-------------------------------------------------------------------

/*******************************************************************\
| Platform Dependent Information Modules                            |
\*******************************************************************/

/*******************************************************************\
| MEM_*.CC Definitions                                              |
\*******************************************************************/

void MEMInit (void);
void MEMGetApplicationMemory (uint64 *pMemoryResident, uint64 *pMemorySwapped);
void MEMGetKernelMemory (uint64 *pMemoryUsed);
void MEMGetTotalMemory (uint64 *pPhysicalUsed, uint64 *pPhysicalFree, uint64 *pSwapUsed);

/*******************************************************************\
| NET_*.CC Definitions                                              |
\*******************************************************************/

void NETInit (void);
void NETGetTotalTraffic (uint64 *pSentBytes, uint64 *pReceivedBytes, uint64 *pSentPackets, uint64 *pReceivedPackets);

/*******************************************************************\
| PRO_*.CC Definitions                                              |
\*******************************************************************/

void PROInit (void);
void PROGetApplicationThreads (uint64 *pThreads);
void PROGetTotalUsage (uint64 *pKernelTicks, uint64 *pWaitTicks, uint64 *pUserTicks, uint64 *pIdleTicks);

//-------------------------------------------------------------------

/*******************************************************************\
| Platform Dependent Library Modules                                |
\*******************************************************************/

/*******************************************************************\
| LOC_*.CC Definitions                                              |
\*******************************************************************/

void LOCAnnounceServer (Server_ptr pServer);
void LOCLocateServer   (Server_out pServer);

/*******************************************************************\
| ORB_*.CC Definitions                                              |
\*******************************************************************/

// pORB			Reference to ORB
// pRootPOA		Reference to root POA
// pRootPOAManager	Reference to manager of root POA

extern CORBA::ORB_ptr                  	pORB;
extern PortableServer::POA_ptr         	pRootPOA;
extern PortableServer::POAManager_ptr	pRootPOAManager;

// ORBInitClient	Initializes broker in the client
// ORBInitServer	Initializes broker in the server
// ORBDoneClient	Shuts down broker in the client
// ORBDoneServer        Shuts down broker in the server

void ORBInitClient (int &iArgC, char *apArgV[]);
void ORBInitServer (int &iArgC, char *apArgV[]);
void ORBDoneClient (void);
void ORBDoneServer (void);

/*******************************************************************\
| RND_*.CC Definitions                                              |
\*******************************************************************/

void   RNDInit      (uint64 iSeed);
uint64 RNDGetRandom (uint64 iModulo = 0);

/*******************************************************************\
| THR_*.CC Definitions                                              |
\*******************************************************************/

void      THRThreadCreate  (THRThread *pThread, void *(*pFunction) (void *), void *pArgument);
void      THRThreadFinish  (THRThread *pThread);
void      THRThreadSleep   (int iSeconds);

void      THRLockCreate    (THRLock *pLock);
void      THRLockFinish    (THRLock *pLock);
void      THRLockLock      (THRLock *pLock);
void      THRLockUnlock    (THRLock *pLock);

/*******************************************************************\
| TIM_*.CC Definitions                                              |
\*******************************************************************/

extern uint64 iTIMTimerScale;
extern uint64 iTIMTimerGranularity;

void   TIMInit         (void);
uint64 TIMGetTimestamp (void);

//-------------------------------------------------------------------

