/*******************************************************************\
| BenchGlobal.java                                                  |
+-------------------------------------------------------------------+
| Definitions and utility methods for all shared modules.           |
+-------------------------------------------------------------------+
| Copyright (C) Lubomir Bulej <lubomir.bulej@mff.cuni.cz>           |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

import java.util.Map;
import java.util.Set;
import java.util.Random;
import java.util.HashSet;
import java.util.Hashtable;


/*******************************************************************\
| BenchGlobal                                                       |
\*******************************************************************/

public final class BenchGlobal {

	/*
	 * General constants.
	 */
	public static final String SAM_BenchmarkName	= "Jampler";
	public static final String SAM_GlobalVersion	= "1.07";
	public static final String SAM_GlobalDate		= "02/12/10";

	/*
	 * Sample parameters.
	 */
	public static int SAM_TestSamples				= 5;

	public static final int SAM_MaxItems			= 100000;
	public static final int SAM_MaxOctets			= 100000;
	public static final int SAM_MaxThreads			= 100;
	public static final int SAM_MaxInstances		= 100000;

	/*
	 * Measurement parameters.
	 *
	 * MEA_WarmUpTimeMin	Minimum time to spend in warm up
	 * MEA_WarmUpTimeMax	Maximum time to spend in warm up
	 * MEA_UsagedCycles		Cycles to spend in usage measurements
	 * MEA_SoloedCycles		Cycles to spend in solo measurements
	 * MEA_SharedTime		Time to spend in shared measurements
	 */
	public static int MEA_WarmUpTimeMin				= 5;
	public static int MEA_WarmUpTimeMax				= 20;
	public static int MEA_UsagedCycles				= 60;
	public static int MEA_SoloedCycles				= 5000;
	public static int MEA_SharedTime				= 60;

	/*
	 * Display definitions.
	 */
	public static final String DSP_HeaderBenchie	= "[BENCHMARK]\t";
	public static final String DSP_HeaderConfig		= "[CONFIGURATION]\t";
	public static final String DSP_HeaderMeasure	= "[MEASUREMENT]\t";
	public static final String DSP_HeaderSample		= "[SAMPLE]\t";

	public static final String DSP_FieldVersion		= "[STR]\tVersion\t";
	public static final String DSP_FieldSuite		= "[STR]\tSuite\t";
	public static final String DSP_FieldDirection	= "[STR]\tDirection\t";
	public static final String DSP_FieldType		= "[STR]\tType\t";
	public static final String DSP_FieldItems		= "[1]\tItems\t";
	public static final String DSP_FieldScale		= "[1]\tScale\t";
	public static final String DSP_FieldGranularity	= "[1]\tGranularity\t";
	public static final String DSP_FieldSamples		= "[1]\tSamples\t";
	public static final String DSP_FieldOctets		= "[1]\tOctets\t";
	public static final String DSP_FieldInstances	= "[1]\tInstances\t";
	public static final String DSP_FieldThreads		= "[1]\tThreads\t";
	public static final String DSP_FieldClients		= "[1]\tClients\t";

	// These are without field length specification
	public static final String DSP_FieldRoundAvg	= "RoundtripAverage";
	public static final String DSP_FieldRoundSolo	= "RoundtripSolo";
	public static final String DSP_FieldThruputAvg	= "ThruputAverage";
	public static final String DSP_FieldThruputSolo	= "ThruputSolo";

	//-------------------------------------------------------------------

	/**
	 * Random number generator.
	 */
	private static Random random;

	/**
	 * Hash sets for verification of argument
	 * passing mode and argument type.
	 */
	private static Set argModes;
	private static Map argTypes;

	static {
		argModes = new HashSet (4);
		argModes.add ("IN");
		argModes.add ("OUT");
		argModes.add ("INOUT");
		argModes.add ("RETURN");

		argTypes = new Hashtable (13);
		argTypes.put ("Short", new Integer (2));
		argTypes.put ("Long", new Integer (4));
		argTypes.put ("LongLong", new Integer (8));
		argTypes.put ("UShort", new Integer (2));
		argTypes.put ("ULong", new Integer (4));
		argTypes.put ("ULongLong", new Integer (8));
		argTypes.put ("Float", new Integer (4));
		argTypes.put ("Double", new Integer (8));
		argTypes.put ("LongDouble", new Integer (16));
		argTypes.put ("Char", new Integer (1));
		argTypes.put ("WChar", new Integer (2));
		argTypes.put ("Boolean", new Integer (1));
		argTypes.put ("Octet", new Integer (1));

		random = new Random (SysInformer.timGetTimestamp ());

		SAM_TestSamples		= getIntProperty ("jampler.test.samples", SAM_TestSamples);
		MEA_SharedTime		= getIntProperty ("jampler.shared.time", MEA_SharedTime);
		MEA_UsagedCycles	= getIntProperty ("jampler.usage.time", MEA_UsagedCycles);
		MEA_SoloedCycles	= getIntProperty ("jampler.solo.cycles", MEA_SoloedCycles);
	}

	//-------------------------------------------------------------------

	/**
	 * Displays benchmark header, including version and suite name.
	 */
	public static void displayHeader (String name) {
		System.out.println (
			DSP_HeaderBenchie +
			DSP_FieldVersion +
				SAM_BenchmarkName +" "+
				SAM_GlobalVersion +" "+
				SAM_GlobalDate +"\t"+
			DSP_FieldSuite + name + "\t");
	} // displayHeader (String)


	/**
	 * Displays benchmark configuration.
	 */
	public static void displayConfig (long scale, long gran) {
		System.out.println (
			DSP_HeaderConfig +
			DSP_FieldScale + scale +"\t"+
			DSP_FieldGranularity + gran +"\t");
	} // displayConfig (String)


	/**
	 * Displays message explaining fatal error and exits.
	 */
	public static void displayFatal (String cls, String msg) {
		System.err.println ("["+ cls +"] "+ msg);
		System.exit (1);
	} // displayFatal (String, String)


	/**
	 * Displays message explaining fatal error, prints
	 * the Throwable's strack trace and exits.
	 */
	public static void displayFatal (String cls, String msg, Throwable thr) {
		System.err.println ("["+ cls +"] "+ msg);
		thr.printStackTrace ();
		System.exit (1);
	} // displayFatal (String, String, Throwable)

	//-------------------------------------------------------------------

	/**
	 * Returns a random positive number (possibly modulo @modulo).
	 */
	public static long rand (long modulo) {
		long result = Math.abs (random.nextLong ());
		return (modulo != 0) ? result % modulo : result;
	} // rand (long)

	//-------------------------------------------------------------------

	/**
	 * Retrieves a positive integer property with default value.
	 */
	public static int getIntProperty (String name, int dval)
		throws IllegalArgumentException {

		try {
			int val = Integer.getInteger (name, dval).intValue ();
			if (val <= 0) {
				throw new IllegalArgumentException ();
			}

			return val;

		} catch (IllegalArgumentException ex) {
			throw new IllegalArgumentException ("Value of "+ name +" is not a positive number!");
		}
	} // getIntProperty (String, int)


	/**
	 * Retrieves a string property indicating argument type.
	 */
	public static String getArgType ()
		throws IllegalArgumentException {

		String value = System.getProperty ("jampler.argument.type");
		if (value == null) {
			throw new IllegalArgumentException ("Argument type not defined!");
		}

		if (! argTypes.containsKey (value)) {
			throw new IllegalArgumentException ("Invalid argument type! ("+ value +")");
		}

		return value;
	} // getArgType ()


	/**
	 * Returns the size of argument type in bytes.
	 */
	public static int getArgSize ()
		throws IllegalArgumentException {

		String value = System.getProperty ("jampler.argument.type");
		if (value == null) {
			throw new IllegalArgumentException ("Argument type not defined!");
		}

		Integer size = (Integer) argTypes.get (value);
		if (size == null) {
			throw new IllegalArgumentException ("Invalid argument type! ("+ value +")");
		}

		return size.intValue ();
	} // getArgSize ()


	/**
	 * Retrieves a string property indicating
	 * argument passing direction.
	 */
	public static String getArgMode ()
		throws IllegalArgumentException {

		String value = System.getProperty ("jampler.argument.mode");
		if (value == null) {
			throw new IllegalArgumentException ("Argument passing mode not defined!");
		}

		if (! argModes.contains (value)) {
			throw new IllegalArgumentException ("Invalid argument passing mode! ("+ value +")");
		}

		return value;
	} // getArgMode ()

} // public final class BenchGlobal
