/*******************************************************************\
| MeasureCore.java                                                  |
+-------------------------------------------------------------------+
| Core measurement routines.                                        |
+-------------------------------------------------------------------+
| Copyright (C) Lubomir Bulej <lubomir.bulej@mff.cuni.cz>           |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

import org.omg.CORBA.IntHolder;
import org.omg.CORBA.LongHolder;
import org.omg.CORBA.SystemException;


/*******************************************************************\
| MeasureCore                                                       |
\*******************************************************************/

public final class MeasureCore {
	/**
	 * Maximal number of parallel threads.
	 */
	public static final int MEA_MaxThreads	= 1024;

	/**
	 * Measurement phases configuration.
	 */
	public static boolean MEA_DoWarmUp 		= true;
	public static boolean MEA_DoServerUsage = true;
	public static boolean MEA_DoClientUsage = true;
	public static boolean MEA_DoSoloRun		= true;
	public static boolean MEA_DoSharedRun	= true;

	static {
		String mode = System.getProperty ("jampler.measure.mode", "default");
		if (mode.equals ("default")) {
			// default

		} else if (mode.equals ("soloonly")) {
			MEA_DoWarmUp 		= false;
			MEA_DoServerUsage	= false;
			MEA_DoClientUsage	= false;
			MEA_DoSharedRun		= false;

		} else if (mode.equals ("sharedonly")) {
			MEA_DoWarmUp		= false;
			MEA_DoServerUsage	= false;
			MEA_DoClientUsage	= false;
			MEA_DoSoloRun		= false;

		} else if (mode.equals ("nousage")) {
			MEA_DoServerUsage	= false;
			MEA_DoClientUsage	= false;

		} else if (mode.equals ("sharednousage")) {
			MEA_DoServerUsage	= false;
			MEA_DoClientUsage	= false;
			MEA_DoSoloRun		= false;

		} else {
			BenchGlobal.displayFatal ("MEA", "Invalid measurement mode ("+ mode +")");
		}
	}

	/*
	 * Indices into array of min/avg/max measurements.
	 */
	private static final int MEA_ApplicationResident	= 0;
	private static final int MEA_ApplicationSwapped		= 1;

	private static final int MEA_KernelUsed 			= 2;

	private static final int MEA_TotalPhysicalUsed 		= 3;
	private static final int MEA_TotalPhysicalFree 		= 4;
	private static final int MEA_TotalSwapUsed	   		= 5;

	private static final int MEA_SentBytes		  		= 6;
	private static final int MEA_ReceivedBytes	   		= 7;
	private static final int MEA_SentPackets			= 8;
	private static final int MEA_ReceivedPackets   		= 9;

	private static final int MEA_ApplicationThreads		= 10;

	private static final int MEA_TotalKernel			= 11;
	private static final int MEA_TotalWait		  		= 12;
	private static final int MEA_TotalUser		   		= 13;
	private static final int MEA_TotalIdle		   		= 14;

	private static final int MEA_NumSamples				= 15;


	/**
	 * Names of measured samples.
	 *
	 * This is required for printing of measured
	 * sample values in a loop.
	 */
	private static final String [] sampleNames = {
		"MemoryApplicationResident",
		"MemoryApplicationSwapped",

		"MemoryKernelUsed",

		"MemoryTotalPhysicalUsed",
		"MemoryTotalPhysicalFree",
		"MemoryTotalSwapUsed",

		"NetworkBytesSent",
		"NetworkBytesReceived",
		"NetworkPacketsSent",
		"NetworkPacketsReceived",

		"ProcessorApplicationThreads",

		"ProcessorTotalKernel",
		"ProcessorTotalWait",
		"ProcessorTotalUser",
		"ProcessorTotalIdle"
	};


	/**
	 * Indices to arrays of counter samples.
	 *
	 * This is required for processing of measured
	 * sample values in a loop.
	 */
	private static final int [] sampleCounters = {
		SysInformer.MIDX_MemoryResident,
		SysInformer.MIDX_MemorySwapped,

		SysInformer.MIDX_MemoryUsed,

		SysInformer.MIDX_PhysicalUsed,
		SysInformer.MIDX_PhysicalFree,
		SysInformer.MIDX_SwapUsed,

		SysInformer.NIDX_SentBytes,
		SysInformer.NIDX_ReceivedBytes,
		SysInformer.NIDX_SentPackets,
		SysInformer.NIDX_ReceivedPackets,

		SysInformer.PIDX_Threads,

		SysInformer.PIDX_KernelTicks,
		SysInformer.PIDX_WaitTicks,
		SysInformer.PIDX_UserTicks,
		SysInformer.PIDX_IdleTicks
	};

	//-------------------------------------------------------------------

	/**
	 * Takes a snapshot of resource usage counters.
	 */
	public static void usage () {
		/*
		 * Sample min/avg/max holders.
		 */
		LongSampler [] samples 			= new LongSampler [MEA_NumSamples];

		/*
		 * Variables related to memory usage.
		 */
		long [] memApplicationMemory 	= new long [2];
		long [] memKernelMemory 		= new long [1];
		long [] memTotalMemory 			= new long [3];

		/*
		 * Variables related to network usage.
		 */
		long [] netTotalTrafficBefore 	= new long [4];
		long [] netTotalTrafficAfter 	= new long [4];
		long [] netTotalTrafficDelta	= new long [4];

		/*
		 * Variables related to processor usage.
		 */
		long [] proApplicationThreads	= new long [1];

		long [] proTotalUsageBefore 	= new long [4];
		long [] proTotalUsageAfter 		= new long [4];
		long [] proTotalUsageDelta		= new long [4];

		long proTotalUsage;

		/*
		 * Variables related to time measurement.
		 */
		long timeBefore;
		long timeAfter;
		long timeDelta;

		//-----------------------------------------------------------

		for (int snum = 0; snum < MEA_NumSamples; snum++) {
			samples [snum] = new LongSampler ();
		}

		/*
		 * The snapshot of time is always taken before the snapshot
		 * of other system information to compensate for the time
		 * spent inside the routines that query this information.
		 */
		timeBefore = SysInformer.timGetTimestamp ();

		SysInformer.netGetTotalTraffic (netTotalTrafficBefore);
		SysInformer.proGetTotalUsage (proTotalUsageBefore);

		for (int mnum = 0; mnum < BenchGlobal.MEA_UsagedCycles; mnum++) {
			sleepWell (1);

			timeAfter = SysInformer.timGetTimestamp ();
			timeDelta = timeAfter - timeBefore;
			timeBefore = timeAfter;

			/*
			 * Take snapshot of the counters.
			 */
			SysInformer.memGetApplicationMemory (memApplicationMemory);
			SysInformer.memGetKernelMemory (memKernelMemory);
			SysInformer.memGetTotalMemory (memTotalMemory);

			SysInformer.netGetTotalTraffic (netTotalTrafficAfter);

			SysInformer.proGetApplicationThreads (proApplicationThreads);
			SysInformer.proGetTotalUsage (proTotalUsageAfter);

			/*
			 * Update absolute memory usage and thread count samples.
			 */
			updateSamples (samples, memApplicationMemory, MEA_ApplicationResident, MEA_ApplicationSwapped);
			updateSamples (samples, memKernelMemory, MEA_KernelUsed);
			updateSamples (samples, memTotalMemory, MEA_TotalPhysicalUsed, MEA_TotalSwapUsed);

			updateSamples (samples, proApplicationThreads, MEA_ApplicationThreads);

			/*
			 * Update network transmission speed samples.
			 */
			arrayDiffSum (netTotalTrafficAfter, netTotalTrafficBefore, netTotalTrafficDelta);
			System.arraycopy (netTotalTrafficAfter, 0, netTotalTrafficBefore, 0, netTotalTrafficBefore.length);
			for (int cnum = 0; cnum < netTotalTrafficDelta.length; cnum++) {
				netTotalTrafficDelta [cnum] *= SysInformer.timTimerScale;
				netTotalTrafficDelta [cnum] /= timeDelta;
			}

			updateSamples (samples, netTotalTrafficDelta, MEA_SentBytes, MEA_ReceivedPackets);

			/*
			 * Update processor usage samples.
			 */
			proTotalUsage = arrayDiffSum (proTotalUsageAfter, proTotalUsageBefore, proTotalUsageDelta);
			System.arraycopy (proTotalUsageAfter, 0, proTotalUsageBefore, 0, proTotalUsageBefore.length);
			if (proTotalUsage != 0) {
				for (int cnum = 0; cnum < proTotalUsageDelta.length; cnum++) {
					proTotalUsageDelta [cnum] *= 100;
					proTotalUsageDelta [cnum] /= proTotalUsage;
				}

				updateSamples (samples, proTotalUsageDelta, MEA_TotalKernel, MEA_TotalIdle);
			}
		} // measurement loop

		/*
		 * Normalize all monitored values and print them out.
		 */
		for (int snum = 0; snum < samples.length; snum++) {
			samples [snum].normalize (BenchGlobal.MEA_UsagedCycles);
			samples [snum].print (sampleNames [snum]);
		}
	} // usage ()


	/**
	 * Measures repeated execution of the suite client action.
	 *
	 * @param name		name of the measured action
	 * @param stamps	descriptions of stamps returned by the action
	 * @param threads	number of threads to start
	 * @param samples	number of samples to make
	 */
	public static void measure (
		String name, ActionStamp stamps [], int threads, int samples) {

		/*
		 * Thread pools, suite meter, shared arguments.
		 */
		MeasureArgs		measureArgs = new MeasureArgs (stamps);
		SuiteMeter		suiteMeter = new SuiteMeter (measureArgs);

		Thread [] 		suiteThreads = new Thread [MEA_MaxThreads];

		/*
		 * Timer variables, measurement counters.
		 */
		long 			timeBefore, timeAfter;
		LongHolder		timeHalfway = new LongHolder ();

		LongHolder		timerScale = new LongHolder ();
		LongHolder		timerGranularity = new LongHolder ();

		int				thread, sample;
		String 			line;

		//-----------------------------------------------------------

		if (threads > MEA_MaxThreads) {
			BenchGlobal.displayFatal ("BEN", "Client Too many threads specified!");
		}

		/*
		 * The measurement.
		 */
		try {
			// Display measurement header on both server and client.
			line = BenchGlobal.DSP_HeaderMeasure +
				BenchGlobal.DSP_FieldSamples +
				samples +"\t"+ name +"\n";

			System.out.print (line);
			BenchClient.benchServer.PrintString (line);

			/*
			 * Request for recording server absolute stamps will be overriden
			 * when the server clock is out of sync with the local clock.
			 */
			timeBefore = SysInformer.timGetTimestamp ();
			BenchClient.benchServer.TimerQuery (timeHalfway, timerScale, timerGranularity);
			timeAfter = SysInformer.timGetTimestamp ();

			if ((timerScale.value != SysInformer.timTimerScale) ||
				(timeBefore > timeHalfway.value) ||
				(timeAfter < timeHalfway.value)) {

				for (int snum = 0; snum < stamps.length; snum++) {
					stamps [snum].flags &= ~ActionStamp.ASF_Remote;
				}
			}

			/*
			 * Initialize suite meter thread arguments.
			 */
			for (sample = 0; sample < samples; sample++) {
				// Display sample header on both server and client.
				System.out.print (BenchGlobal.DSP_HeaderSample);
				BenchClient.benchServer.PrintString (BenchGlobal.DSP_HeaderSample);

				/*
				 * Create and start measurement threads to do the
				 * measuring and wait for all of them to initialize.
				 */
				measureArgs.threadsToRun	= threads;
				measureArgs.threadPrincipal = (int) BenchGlobal.rand (threads);
				measureArgs.doSoloedBeg 	= false;
				measureArgs.doSharedBeg 	= false;
				measureArgs.doSharedEnd 	= false;
				measureArgs.totalCount 		= 0;

				measureArgs.avgRoundtrip.initialize ();
				for (int snum = 0; snum < stamps.length; snum++) {
					measureArgs.avgTimestamps [snum].initialize ();
				}

				for (thread = 0; thread < threads; thread++) {
					suiteThreads [thread] = new Thread (suiteMeter);
					suiteThreads [thread].start ();
				}

				measureArgs.allThreadsRunning.down ();

				// PHASE 1: WARMUP

				SysInformer.memGarbageCollect ();

				if (MEA_DoWarmUp) {
					sleepWell (BenchGlobal.MEA_WarmUpTimeMin + BenchGlobal.rand (
						BenchGlobal.MEA_WarmUpTimeMax - BenchGlobal.MEA_WarmUpTimeMin));
				}

				// PHASE 2: USAGE CLIENT

				if (MEA_DoClientUsage) {
					usage ();
				}

				// PHASE 3: USAGE SERVER

				if (MEA_DoServerUsage) {
					BenchClient.benchServer.UsageStart ();
					sleepWell (BenchGlobal.MEA_UsagedCycles);
					do {
						sleepWell (1);
					} while (BenchClient.benchServer.UsageQuery ());
				}

				// PHASE 4: SOLO RUN

				if (MEA_DoSoloRun) {
					measureArgs.doSoloedBeg = true;
					measureArgs.soloThreadFinished.down ();
				}

				// PHASE 5: SHARED RUN

				if (MEA_DoSharedRun) {
					timeBefore = SysInformer.timGetTimestamp ();
					measureArgs.doSharedBeg = true;

					sleepWell (BenchGlobal.MEA_SharedTime);

					timeAfter = SysInformer.timGetTimestamp ();
					measureArgs.doSharedEnd = true;
				}

				/*
				 * Wait for the threads to finish before
				 * doing anything else.
				 */
				for (thread = 0; thread < threads; thread++) {
					joinThread (suiteThreads [thread]);
				}

				// Display the rest of output.
				if (MEA_DoSharedRun) {
					long deltaTime = timeAfter - timeBefore;

					// average thruput
					System.out.print ("[1]\t"+ BenchGlobal.DSP_FieldThruputAvg +"\t");
					System.out.print ((measureArgs.totalCount * SysInformer.timTimerScale) / deltaTime +"\t");

					// solo thruput
					System.out.print ("["+ threads +"]\t"+ BenchGlobal.DSP_FieldThruputSolo +"\t");
					for (int tnum = 0; tnum < threads; tnum++) {
						System.out.print ((measureArgs.threadCounts [tnum] * SysInformer.timTimerScale) / deltaTime +"\t");
					}
				}

				// Display timestamp data
				for (int snum = 0; snum < stamps.length; snum++) {
					if (((stamps [snum].flags &
						(ActionStamp.ASF_Local | ActionStamp.ASF_Remote)) != 0)) {

						display (
							stamps [snum].name + "Average",
							measureArgs.avgTimestamps [snum],
							measureArgs.totalCount,
							stamps [snum].name + "Solo",
							measureArgs.soloTimestamps [snum]);
					}
				}

				// Display roundtrip data
				display (
					BenchGlobal.DSP_FieldRoundAvg,
					measureArgs.avgRoundtrip,
					measureArgs.totalCount,
					BenchGlobal.DSP_FieldRoundSolo,
					measureArgs.soloRoundtrip);

				// The last line
				System.out.println ();
				BenchClient.benchServer.PrintString ("\n");
			} // sample loop

		} catch (SystemException ex) {
			BenchGlobal.displayFatal ("MEA", "Thread "+ ex.toString ());
		} catch (Exception ex) {
			BenchGlobal.displayFatal ("MEA", "Thread "+ ex.getMessage (), ex);
		}
	} // measure (int, String, int, int)

	//-------------------------------------------------------------------

	public static void initialStep (
		IntHolder step, IntHolder current, int min, int scale) {

		step.value = 1;
		current.value = min;
		while (step.value < min && scale > 1) {
			step.value *= scale;
		}
	} // initialStep (IntHolder, IntHolder, int, int)

	public static void advanceStep (
		IntHolder step, IntHolder current, int limit, int scale) {

		if (current.value <= step.value * limit || scale == 1) {
			current.value += step.value;
		} else {
			step.value *= scale;
			current.value = step.value;
		}
	} // advanceStep (IntHolder, IntHolder, int, int)

	//-------------------------------------------------------------------

	/**
	 * Try-hard wrappers to various interruptible tasks.
	 */
	public static void sleepWell (long secs) {
		while (true) {
			try {
				Thread.sleep (1000 * secs);
				break;
			} catch (InterruptedException ex) { /* nop */ }
		}
	} // sleepWell (long)

	public static void joinThread (Thread thread) {
		while (true) {
			try {
				thread.join ();
				break;
			} catch (InterruptedException ex) { /* nop */ }
		}
	} // joinThread (Thread)

	//-------------------------------------------------------------------

	/**
	 * Displays measurement data for all threads and the solo thread.
	 */
	private static void display (
		String avgHead, LongSampler avg, long avgNorm,
		String soloHead, long [] solo) {

		if (MEA_DoSharedRun) {
			avg.normalize (avgNorm);
			avg.print (avgHead);
		}

		if (MEA_DoSoloRun) {
			System.out.print ("["+ solo.length +"]\t"+ soloHead +"\t");
			for (int mnum = 0; mnum < solo.length; mnum++) {
				System.out.print (solo [mnum] +"\t");
			}
		}
	} // display (String, LongSampler, long, String, long [])


	/**
	 * Computes differences between elements of two arrays into
	 * @delta array and returns the sum of the differences. Both
	 * arrays should be of equal length and long at least as
	 * the @delta array.
	 */
	public static long arrayDiffSum (
		long [] left, long [] right, long [] delta) {

		long sum = 0;
		for (int idx = 0; idx < delta.length; idx++) {
			delta [idx] = left [idx] - right [idx];
			sum += delta [idx];
		}

		return sum;
	} // arrayDiffSum (long [], long [], long [])


	/**
	 * Updates measured samples with given values. Indices into
	 * @data array are computed by mapping sample index through
	 * the @sampleCounters array.
	 */
	private static void updateSamples (
		LongSampler [] samples, long [] data, int beg, int end) {

		for (/* nop */; beg <= end; beg++) {
			samples [beg].update (data [sampleCounters [beg]]);
		}
	} // updateSamples (LongSampler [], long [], int, int)

	private static void updateSamples (
		LongSampler [] samples, long [] data, int snum) {

		samples [snum].update (data [sampleCounters [snum]]);
	} // updateSamples (LongSampler [], long [], int)

} // public final class MeasureCore


/*******************************************************************\
| Semaphore                                                         |
\*******************************************************************/

/**
 * Simple semaphore for coordination between
 * benchmark control and measurement threads.
 */
final class Semaphore {

	private int avail = 0;


	public synchronized final void down () {
		avail--;

		if (avail < 0) {
			while (true) {
				try {
					wait ();
					break;
				} catch (InterruptedException ex) { /* nop */ }
			}
		}
	} // down ()


	public synchronized final void up () {
		avail++;

		if (avail <= 0) {
			notify ();
		}
	} // up ()

} // final class Semaphore


/*******************************************************************\
| MeasureArgs                                                       |
\*******************************************************************/

/**
 * Measurement data shared by all SuiteMeter threads.
 */
final class MeasureArgs {
	ActionStamp [] stamps			= null;

	/*
	 * Shared data on threads.
	 */
	int	threadsToRun				= 0;
	int threadPrincipal				= 0;

	boolean doSoloedBeg				= false;
	boolean doSharedBeg				= false;
	boolean doSharedEnd 			= false;

	Semaphore allThreadsRunning 	= null;
	Semaphore soloThreadFinished	= null;

	/*
	 * Shared data on measurements.
	 */
	long totalCount					= 0;
	long [] threadCounts			= null;
	long [] soloRoundtrip 			= null;
	long [] [] soloTimestamps		= null;

	LongSampler avgRoundtrip		= null;
	LongSampler	[] avgTimestamps	= null;


	/**
	 * Initializes dynamically allocated fields.
	 */
	public MeasureArgs (ActionStamp [] stamps) {
		this.stamps = stamps;

		threadCounts = new long [MeasureCore.MEA_MaxThreads];

		allThreadsRunning = new Semaphore ();
		soloThreadFinished = new Semaphore ();

		soloTimestamps = new long [stamps.length] [];
		for (int snum = 0; snum < soloTimestamps.length; snum++) {
			soloTimestamps [snum] = new long [BenchGlobal.MEA_SoloedCycles];
		}

		soloRoundtrip = new long [BenchGlobal.MEA_SoloedCycles];

		avgTimestamps = new LongSampler [stamps.length];
		for (int snum = 0; snum < avgTimestamps.length; snum++) {
			avgTimestamps [snum] = new LongSampler ();
		}

		avgRoundtrip = new LongSampler ();
	} // MeasureArgs (ActionStamp)

} // final class MeasureArgs


/*******************************************************************\
| SuiteMeter                                                        |
\*******************************************************************/

/**
 * Measurement thread performing the actual benchmarking. When a thread
 * invokes the run() method on the SuiteMeter, it allocates separate
 * instances of SuiteAction so that its instance variables are local to
 * the executing thread.
 */
final class SuiteMeter implements Runnable {

	private MeasureArgs args;


	public SuiteMeter (MeasureArgs args) {
		this.args = args;
	} // SuiteMeter (MeasureArgs)


	public final void run () {
		long []			threadTimes;
		LongSampler		avgRoundtrip;
		LongSampler []	avgTimestamps;

		SuiteAction		action = new SuiteAction ();

		/*
		 * Initialize the locals.
		 */
		threadTimes = new long [args.stamps.length];

		avgTimestamps = new LongSampler [args.stamps.length];
		for (int snum = 0; snum < avgTimestamps.length; snum++) {
			avgTimestamps [snum] = new LongSampler ();
		}

		avgRoundtrip = new LongSampler ();


		/*
		 * Do the per-thread measurement.
		 */
		try {
			int		threadNumber;
			long 	timeBefore;
			long	timeAfter;

			/*
			 * Determine the thread number and decrement the number
			 * of threads to run in the benchmark session. The last
			 * thread to execute will notify the coordination routine
			 * so that it can start the measurement and all.
			 */
			synchronized (args) {
				args.threadsToRun--;
				threadNumber = args.threadsToRun;
			}

			if (threadNumber == 0) {
				args.allThreadsRunning.up ();
			}

			//-----------------------------------------------------------

			/*
			 * It is only the principal thread that collects data in the
			 * solo phase of the benchmark, because otherwise it would
			 * simply be too much data.
			 */
			if (MeasureCore.MEA_DoSoloRun && threadNumber == args.threadPrincipal) {

				// wait for the coordinator
				while (! args.doSoloedBeg) {
					action.execute (threadTimes);
				}

				// collect the data
				for (int mnum = 0; mnum < BenchGlobal.MEA_SoloedCycles; mnum++) {
					timeBefore = SysInformer.timGetTimestamp ();
					action.execute (threadTimes);
					timeAfter = SysInformer.timGetTimestamp ();

					// convert and copy timestamps
					for (int snum = 0; snum < threadTimes.length; snum++) {
						int flags = args.stamps [snum].flags;

						args.soloTimestamps [snum] [mnum] =
							(flags & ActionStamp.ASF_Absolute) != 0 ?
							threadTimes [snum] - timeBefore :
							threadTimes [snum];
					}

					args.soloRoundtrip [mnum] = timeAfter - timeBefore;
				} // solo measurement loop

				// notify the coordinator
				args.soloThreadFinished.up ();
			} // solo stuff

			//-----------------------------------------------------------

			/*
			 * All threads collect timing information until they are
			 * told to stop. The information is then added together.
			 */
			args.threadCounts [threadNumber] = 0;
			if (MeasureCore.MEA_DoSharedRun) {
				while (! args.doSharedBeg) {
					action.execute (threadTimes);
				}

				while (! args.doSharedEnd) {
					timeBefore = SysInformer.timGetTimestamp ();
					action.execute (threadTimes);
					timeAfter = SysInformer.timGetTimestamp ();

					// convert timestamps and update the samplers
					for (int snum = 0; snum < threadTimes.length; snum++) {
						int flags = args.stamps [snum].flags;

						avgTimestamps [snum].update (
								(flags & ActionStamp.ASF_Absolute) != 0 ?
								threadTimes [snum] - timeBefore :
								threadTimes [snum]);
					}

					avgRoundtrip.update (timeAfter - timeBefore);
					args.threadCounts [threadNumber]++;
				} // shared measurement loop
			} // if DoSharedRun

			/*
			 * Update shared measurement data.
			 */
			synchronized (args) {
				for (int snum = 0; snum < avgTimestamps.length; snum++) {
					args.avgTimestamps [snum].update (avgTimestamps [snum]);
				}

				args.avgRoundtrip.update (avgRoundtrip);
				args.totalCount += args.threadCounts [threadNumber];
			}

		} catch (SystemException ex) {
			BenchGlobal.displayFatal ("MEA", "Thread "+ ex.toString ());
		} catch (Exception ex) {
			BenchGlobal.displayFatal ("MEA", "Thread "+ ex.getMessage (), ex);
		}
	} // run ()

} // final class SuiteMeter
