/*******************************************************************\
| PeerLocator.java                                                  |
+-------------------------------------------------------------------+
| Benchmark server locator.                                         |
+-------------------------------------------------------------------+
| Copyright (C) Lubomir Bulej <lubomir.bulej@mff.cuni.cz>           |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.BufferedReader;

import java.util.Stack;

import org.omg.CORBA.UserException;
import org.omg.CORBA.SystemException;

import org.omg.CosNaming.NameComponent;
import org.omg.CosNaming.NamingContext;
import org.omg.CosNaming.NamingContextHelper;


/*******************************************************************\
| PeerLocator                                                       |
\*******************************************************************/

/**
 * The PeerLocator class serves for announcing and locating the benchmark
 * servant against which the measurement should be performed, under name
 * specified in the jampler.locator.name property. The default name is
 * "Jampler". It can either use a file to store and retrieve the benchmark
 * server reference from, or it can use Naming Service for that purpose,
 * which is the default.
 *
 * The contents of the jampler.locator.mode property select the method to
 * be used for announcing and locating the server. To select either method,
 * use the "name" or "file" modes, or a comma separated list of methods to
 * be attempted sequentially, such as "name,file".
 */
public final class PeerLocator {

	private static String benchName;
	private static String iorFileName;

	private static NamingContext cosNaming;
	private static NameComponent benchNameNC [];

	private static final int PLM_None 		= 0;
	private static final int PLM_Name		= 1;
	private static final int PLM_File		= 2;


	/**
	 * Retrieves the requested name of the benchmark and
	 * sets up the components depending on its name.
	 */
	static {
		benchName = System.getProperty ("jampler.locator.name",
			BenchGlobal.SAM_BenchmarkName);

		iorFileName = System.getProperty ("jampler.locator.file",
			benchName +".IOR");
	}


	/**
	 * Announces the benchmark server using the methods specified in the
	 * jampler.locator.mode property.
	 */
	public static void announceServer (Server server) {
		StringBuffer modeListBuf = initModeListBuf ();
		Stack errors = new Stack ();
		boolean success = false;

		/*
		 * Announce the server using the specified methods...
		 */
		do {
			int mode = getLocatorMode (modeListBuf);

			try {
				if (mode == PLM_Name) {
					nameAnnounceServer (server);
					success = true;
				} else if (mode == PLM_File) {
					fileAnnounceServer (server);
					success = true;
				} else {
					// there are no more methods to try
					break;
				}

			} catch (Exception ex) {
				errors.push (ex.toString ());
			}
		} while (! success); // announcer loop


		/*
		 * Check the result...
		 */
		if (! success) {
			if (! errors.empty ()) {
				BenchGlobal.displayFatal ("LOC", "announceServer "+ (String) errors.pop ());
			} else {
				BenchGlobal.displayFatal ("LOC", "announceServer Failed to announce the server!");
			}
		}
	} // announceServer (Server)


	/**
	 * Attempts to locate the server using the methods specified in the
	 * jampler.locator.mode property.
	 */
	public static Server locateServer () {
		StringBuffer modeListBuf = initModeListBuf ();
		Stack errors = new Stack ();
		Server server = null;

		/*
		 * Locate the server using the specified methods...
		 */
		do {
			int mode = getLocatorMode (modeListBuf);

			try {
				if (mode == PLM_Name) {
					server = nameLocateServer ();
				} else if (mode == PLM_File) {
					server = fileLocateServer ();
				} else {
					// there are no more methods to try
					break;
				}

			} catch (Exception ex) {
				errors.push (ex.toString ());
			}

		} while (server == null); // locator loop


		/*
		 * Check the result...
		 */
		if (server == null) {
			if (! errors.empty ()) {
				BenchGlobal.displayFatal ("LOC", "locateServer "+ (String) errors.pop ());
			} else {
				BenchGlobal.displayFatal ("LOC", "locateServer Failed to locate the server!");
			}
		}

		return server;
	} // locateServer ()

	//---------------------------------------------------------------


	/**
	 * Strips locator modes from the passed string buffer and returns the
	 * locator mode type to the caller if it is known.
	 */
	private static int getLocatorMode (StringBuffer modeBuf)
		throws IllegalArgumentException {

		while (modeBuf.length () > 0) {
			int end = 0;
			while (end < modeBuf.length () && modeBuf.charAt (end) != ',') {
				end++;
			}

			String val = modeBuf.substring (0, end).trim ();
			if (end < modeBuf.length () && modeBuf.charAt (end) == ',') {
				end++;
			}

			modeBuf.delete (0, end);
			if (val.equals ("name")) {
				return PLM_Name;
			} else if (val.equals ("file")) {
				return PLM_File;
			} else {
				throw new IllegalArgumentException ("Unknown locator mode! ["+ val +"]");
			}
		}

		return PLM_None;
	} // getLocatorMode (StringBuffer)


	/**
	 * Gets the list of locator modes from the system property and
	 * returns it as a string buffer.
	 */
	private static StringBuffer initModeListBuf () {
		StringBuffer modeListBuf = new StringBuffer ();
		String modes = System.getProperty ("jampler.locator.mode", "name");
		modeListBuf.insert (0, modes.trim ());
		return modeListBuf;
	} // initModeListBuf ()

	//---------------------------------------------------------------

	/**
	 * Looks up reference to CORBA Naming Service and initializes the
	 * benchmark name components for looking up/registering the server in
	 * the naming service.
	 */
	private static void initNaming ()
		throws Exception {

		org.omg.CORBA.Object baseNaming = BenchBroker.theORB.resolve_initial_references ("NameService");
		cosNaming = NamingContextHelper.narrow (baseNaming);

		benchNameNC = new NameComponent [1];
		benchNameNC [0] = new NameComponent (benchName, "");
	} // initNaming ()


	/**
	 * Registers the benchmark server in the naming service.
	 */
	private static void nameAnnounceServer (Server server)
		throws Exception {

		initNaming ();
		cosNaming.rebind (benchNameNC, server);
	} // nameAnnounceServer (Server)


	/**
	 * Attempts to look up the reference to benchmark server in
	 * the naming service and narrows it to the proper type.
	 */
	private static Server nameLocateServer ()
		throws Exception {

		initNaming ();
		org.omg.CORBA.Object baseServer = cosNaming.resolve (benchNameNC);
		return ServerHelper.narrow (baseServer);
	} // nameLocateServer ()

	//---------------------------------------------------------------

	/**
	 * Writes the reference to benchmark server to a file.
	 */
	private static void fileAnnounceServer (Server server)
		throws Exception {

		String serverRef = BenchBroker.theORB.object_to_string (server);

		FileWriter fileWriter = new FileWriter (iorFileName);
		PrintWriter printWriter = new PrintWriter (fileWriter);
		printWriter.println (serverRef);
		printWriter.flush ();
		printWriter.close ();
	} // fileAnnounceServer (Server)


	/**
	 * Attempts to read the reference to benchmark server from a file and
	 * narrows it to the proper type.
	 */
	private static Server fileLocateServer ()
		throws Exception {

		FileReader fileReader = new FileReader (iorFileName);
		BufferedReader bufReader = new BufferedReader (fileReader);
		String serverRef = bufReader.readLine ();
		bufReader.close ();

		org.omg.CORBA.Object baseServer = BenchBroker.theORB.string_to_object (serverRef);
		return ServerHelper.narrow (baseServer);
	} // fileLocateServer ()

} // public final class PeerLocator
