/*******************************************************************\
| SysInformer.java                                                  |
+-------------------------------------------------------------------+
| Mostly native code interface to various system internals.         |
+-------------------------------------------------------------------+
| Copyright (C) Lubomir Bulej <lubomir.bulej@mff.cuni.cz>           |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

public final class SysInformer {
	/*
	 * Counter offsets in arrays returned by
	 * native system information library.
	 */
	public static final int MIDX_MemoryResident		= 0;
	public static final int MIDX_MemorySwapped 		= 1;

	public static final int MIDX_MemoryUsed 		= 0;

	public static final int MIDX_PhysicalUsed 		= 0;
	public static final int MIDX_PhysicalFree 		= 1;
	public static final int MIDX_SwapUsed			= 2;

	public static final int NIDX_SentBytes 			= 0;
	public static final int NIDX_ReceivedBytes 		= 1;
	public static final int NIDX_SentPackets 		= 2;
	public static final int NIDX_ReceivedPackets	= 3;

	public static final int PIDX_Threads			= 0;

	public static final int PIDX_KernelTicks 		= 0;
	public static final int PIDX_WaitTicks 			= 1;
	public static final int PIDX_UserTicks 			= 2;
	public static final int PIDX_IdleTicks 			= 3;

	public static final int TIDX_TimerScale			= 0;
	public static final int TIDX_TimerGranularity   = 1;


	/*
	 * System timer configuration.
	 */
	public static long timTimerScale;
	public static long timTimerGranularity;

	/*
	 * Root thread group.
	 */
	private static ThreadGroup rootTG;

	/*
	 * Java runtime.
	 */
	private static Runtime javaRT;


	/**
	 * Static initializer; loads library with implementation
	 * of native methods, retrieves the timer configuration and
	 * finds root thread group for future use.
	 */
	static {
		long [] cntrs = new long [2];

		System.loadLibrary ("SysInformer");

		timGetConfig (cntrs);
		timTimerScale = cntrs [TIDX_TimerScale];
		timTimerGranularity = cntrs [TIDX_TimerGranularity];

		// Determine root thread group
		rootTG = Thread.currentThread ().getThreadGroup ();
		while (rootTG.getParent () != null) {
			rootTG = rootTG.getParent ();
		}

		javaRT = Runtime.getRuntime ();
	}


	/*
	 * Native code methods provided by the system information library.
	 */
	public static native final void memGetApplicationMemory (long [] cntrs);
	public static native final void memGetKernelMemory (long [] cntrs);
	public static native final void memGetTotalMemory (long [] cntrs);

	public static native final void netGetTotalTraffic (long [] cntrs);

	public static native final void proGetTotalUsage (long [] cntrs);

	public static native final void timGetConfig (long [] cntrs);
	public static native final long timGetTimestamp ();


	/**
	 * Determines the number of threads within the JVM.
	 */
	public static void proGetApplicationThreads (long [] cntrs) {
		cntrs [0] = rootTG.activeCount ();
	} // proGetApplicationThreads (long [])


	/**
	 * Computes approximate heap size.
	 */
	public static long memGetHeapSize () {
		return javaRT.totalMemory () - javaRT.freeMemory ();
	} // memGetHeapSize ()


	/**
	 * Performs garbage collection with some more
	 * pressure than just calling gc() once.
	 */
	public static void memGarbageCollect () {
		long heapBefore, heapAfter;
		int	iterations = 0;

		heapAfter = memGetHeapSize ();
		do {
			heapBefore = heapAfter;
			javaRT.runFinalization ();
			javaRT.gc ();

			Thread.currentThread ().yield ();
			heapAfter = memGetHeapSize ();
			iterations++;
		} while (heapAfter < heapBefore && iterations < 10);
	} // memGarbageCollect ()

} // public final class SysInformer
