/*******************************************************************\
| FileExpander.java                                                 |
+-------------------------------------------------------------------+
| Copyright (C) Lubomir Bulej <lubomir.bulej@mff.cuni.cz>           |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

import java.io.*;
import java.util.*;


/*******************************************************************\
| FileExpander                                                      |
\*******************************************************************/

/**
 * Simple front-end to FileTemplate.
 */
public final class FileExpander {

	public static void main (String args []) {
		FileTemplate template;

		if (args.length == 0) {
			System.err.println ("Usage: FileExpander <file> (<token>[=<value>] )*");
			System.exit (1);
		}


		try {
			template = new FileTemplate (args [0]);

			// read symbol values
			Map symbols = new Hashtable ();
			for (int anum = 1; anum < args.length; anum++) {
				// skip empty strings
				if (args [anum].length () == 0) {
					continue;
				}

				// get name & value
				int pos = args [anum].indexOf ('=');
				if (pos == -1) {
					symbols.put (args [anum], "");
					continue;
				}

				String name = args [anum].substring (0, pos);
				if (pos + 1 == args [anum].length ()) {
					symbols.put (name, "");
					continue;
				}

				// store named values
				String value = args [anum].substring (pos + 1);
				symbols.put (name, value);
			} // definition loop


			// expand the template to stdout
			template.flush (symbols, System.out);

		} catch (FileNotFoundException ex) {
			System.err.println ("Error: file "+ args [0] +" does not exist!");
			System.exit (1);
		} catch (IOException ex) {
			System.err.println ("Error: "+ ex.getMessage ());
			System.exit (1);
		}
	} // main (String [])

} // public final class FileExpander


/*******************************************************************\
| FileTemplate                                                      |
\*******************************************************************/

/**
 * FileTemplate class is used for expanding file
 * templates. These are textual files containing special
 * tokens that are supposed to be replaced by some text.
 * FileTemplate class loads this text file and when asked
 * to output it to a stream, it requests that these tokens
 * be replaced by the caller, or that the replacements be
 * provided in form of hash table.
 */
final class FileTemplate {
	private int linePosition = 0;
	private String currentLine = null;
	private Iterator templatePosition;
	private ArrayList template = new ArrayList ();

	private static String TOKEN_PREFIX = "@";
	private static String TOKEN_SUFFIX = "@";


	/**
	 * Initialize template from an array of strings.
	 */
	public FileTemplate (String [] lines) {
		for (int lnum = 0; lnum < lines.length; lnum++) {
			template.add (lines [lnum]);
		}

		reset ();
	} // FileTemplate (String [])


	/**
	 * Initialize template from a list of strings.
	 */
	public FileTemplate (ArrayList list) {
		template = list;
		reset ();
	} // FileTemplate (ArrayList)


	/**
	 * Initialize template from InputStream.
	 */
	public FileTemplate (InputStream stream)
		throws IOException {

		read (stream);
		reset ();
	} // FileTemplate (InputStream)


	/**
	 * Initialize template from file named @name.
	 */
	public FileTemplate (String name)
		throws IOException {

		FileInputStream stream = new FileInputStream (name);
		read (stream);
		reset ();
	} // FileTemplate (String)


	/**
	 * Read strings from input stream and add them to
	 * the list of template lines.
	 */
	public final void read (InputStream input)
		throws IOException {

		BufferedReader reader = new BufferedReader (
			new InputStreamReader (input));

		try {
			String line = reader.readLine ();
			while (line != null) {
				template.add (line);
				line = reader.readLine ();
			}

		} finally {
			reader.close ();
		}
	} // read (InputStream)


	/**
	 * Reset the position in template to the first line.
	 */
	public final void reset () {
		linePosition = 0;
		templatePosition = template.iterator ();
		currentLine = templatePosition.hasNext () ? (String) templatePosition.next () : null;
	} // reset ()


	public final String expand (PrintStream output) {

		while (currentLine != null) {
			// anything left on the current line?
			if (linePosition < currentLine.length ()) {

				// look for token prefix
				int prefixIndex = currentLine.indexOf (TOKEN_PREFIX, linePosition);
				if (prefixIndex >= 0) {
					// look for the suffix
					int suffixIndex = currentLine.indexOf (TOKEN_SUFFIX, prefixIndex + TOKEN_PREFIX.length ());
					if (suffixIndex > 0) {
						String token = currentLine.substring (prefixIndex + TOKEN_PREFIX.length (), suffixIndex);
						output.print (currentLine.substring (linePosition, prefixIndex));
						linePosition = suffixIndex + TOKEN_SUFFIX.length ();
						return token;
					} // suffix found
				} // prefix found

				// no token found, output rest of the line
				output.print (currentLine.substring (linePosition));
				linePosition = currentLine.length ();

			} else {
				// advance to next line
				currentLine = templatePosition.hasNext () ? (String) templatePosition.next () : null;
				output.println ();
				linePosition = 0;
			}

		} // template loop

		reset ();
		return null;
	} // expand (PrintStream)


	/**
	 * Clear the template.
	 */
	public final void clear () {
		template.clear ();
		reset ();
	} // clear ()


	public final boolean isEmpty () {
		return (template.size () == 0);
	} // isEmpty ()


	public final void map (String token, Map mapping, PrintStream output) {
		String value = (String) mapping.get (token);
		if (value != null) {
			output.print (value);
		} else {
			output.print (TOKEN_PREFIX+ token +TOKEN_SUFFIX);
		}
	} // map (String, HashMap, PrintStream0


	public final void flush (PrintStream output)
		throws IOException {

		while (expand (output) != null)
			;
	} // flush (PrintStream)


	public final void flush (Map mapping, PrintStream output)
		throws IOException {

		String token = expand (output);
		while (token != null) {
			map (token, mapping, output);
			token = expand (output);
		}
	} // flush (Hashtable, PrintStream)

} // final class FileTemplate


