/*******************************************************************\
| FileStripper.java                                                 |
+-------------------------------------------------------------------+
| Copyright (C) Lubomir Bulej <lubomir.bulej@mff.cuni.cz>           |
|               and other contributing authors of                   |
|               Distributed Systems Research Group                  |
|               Department of Software Engineering                  |
|               Faculty of Mathematics and Physics                  |
|               Charles University, Czech Republic                  |
\*******************************************************************/

import java.io.*;
import java.util.*;


/*******************************************************************\
| FileStripper                                                      |
\*******************************************************************/

/*
 * Simple front-end to Preprocessor.
 */
public final class FileStripper {

	public static void main (String args []) {
		Preprocessor preprocessor;

		if (args.length == 0) {
			System.err.println ("Usage: FileStripper <file> <symbol>*");
			System.exit (1);
		}


		try {
			preprocessor = new Preprocessor (args [0]);

			// read symbol values
			Map symbols = new Hashtable ();
			for (int anum = 1; anum < args.length; anum++) {
				// skip empty strings
				if (args [anum].length () == 0) {
					continue;
				}

				// get name & value
				int pos = args [anum].indexOf ('=');
				if (pos == -1) {
					symbols.put (args [anum], "");
					continue;
				}

				String name = args [anum].substring (0, pos);
				if (pos + 1 == args [anum].length ()) {
					symbols.put (name, "");
					continue;
				}

				// store named values
				String value = args [anum].substring (pos + 1);
				symbols.put (name, value);
			} // definition loop

			// preprocess the file
			preprocessor.process (symbols, System.out);

		} catch (FileNotFoundException ex) {
			System.err.println ("Error: file "+ args [0] +" does not exist!");
			System.exit (1);
		} catch (IOException ex) {
			System.err.println ("Error: "+ ex.getMessage ());
			System.exit (1);
		}
	} // main (String [])

} // public final class FileStripper


/*******************************************************************\
| Preprocessor                                                      |
\*******************************************************************/

/**
 * Quick'n'dirty trivial "preprocessor" that can handle #ifdef,
 * #ifndef, #elifdef, #else, #endif and #error directives and
 * does not care about comments.
 *
 * It will probably break given bad enough look...
 */
final class Preprocessor {

	private BufferedReader reader;

	private static final int DIR_INVALID	= -1;
	private static final int DIR_NONE		=  0;
	private static final int DIR_IFDEF		=  1;
	private static final int DIR_IFNDEF		=  2;
	private static final int DIR_ELIFDEF	=  3;
	private static final int DIR_ELSE		=  4;
	private static final int DIR_ENDIF		=  5;
	private static final int DIR_ERROR		=  6;


	public Preprocessor (String name)
		throws IOException {

		reader = new BufferedReader (new FileReader (name));
	} // Preprocessor (name)


	public final void process (Map symbols, PrintStream output)
		throws IOException {

		try {
			int linenr = 1;
			int suppress = 0;
			Stack stack = new Stack ();

			/*
			 * The whole thing works so that it keeps track of number of
			 * negative "expression" results on the stack and suppresses
			 * the output if there are any present...
			 *
			 * Expression on the same level can be active only once, thus
			 * value of 0 on the stack represents "initial negative" result,
			 * value of 1 "once positive" result and all other values
			 * represent "forced negative" result.
			 */
			String orig = reader.readLine ();
			while (orig != null) {
				int result = 0;
				boolean active;

				StringHolder line = new StringHolder (orig.trim ());
				int dir = checkDirective (line);

				switch (dir) {
				case DIR_INVALID:
					fatal ("error:"+ linenr +": invalid directive "+ line.string);
					break;

				case DIR_ELSE:
				case DIR_ENDIF:
				case DIR_ELIFDEF:
					if (stack.empty ()) {
						fatal ("error:"+ linenr +": unmatched directive "+ line.string);
					}

					// restore last result (that's all on #endif)
					result = ((Integer) stack.pop ()).intValue ();
					suppress -= (result == 1) ? 0 : 1;

					// put back "inverse" or "forced negative" value on #else
					if (dir == DIR_ELSE) {
						result++;
						stack.push (new Integer (result));
						suppress += (result == 1) ? 0 : 1;
					}

					if (dir != DIR_ELIFDEF) {
						break;
					}

					// reevaluate on #elifdef: Obfuscated Java Contest entry

				case DIR_IFDEF:
				case DIR_IFNDEF:
					result = (dir == DIR_ELIFDEF) ? result : 0;
					active = (dir == DIR_IFNDEF) ^ symbols.containsKey (line.string);
					result += result + (active ? 1 : 0);
					stack.push (new Integer (result));
					suppress += (result == 1) ? 0 : 1;
					break;

				case DIR_ERROR:
					if (suppress == 0) {
						fatal ("error:"+ linenr +": user error: "+ line.string.trim ());
					}
					break;

				default:
					if (suppress == 0) {
						output.println (orig);
					}
				} // directive switch

				orig = reader.readLine ();
				linenr++;
			}

		} finally {
			reader.close ();
		}
	} // process (Map, PrintStream)


	public final int skipWhitespace (StringHolder line, int pos) {
		while (pos < line.length ()
			&&  Character.isWhitespace (line.charAt (pos))) {

			pos++;
		}

		return pos;
	} // skipWhitespace (StringHolder, int)


	public final int skipLetters (StringHolder line, int pos) {
		while (pos < line.length ()
			&& Character.isLetter (line.charAt (pos))) {

			pos++;
		}

		return pos;
	} // skipLetters (StringHolder, int)


	public final int getWord (StringHolder line, int pos) {
		int beg = skipWhitespace (line, pos);
		int end = skipLetters (line, beg + 1);
		line = line.substring (beg, end);
		return end;
	} // getWord (StringHolder, int)


	public final int checkDirective (StringHolder line) {
		if (line.length () == 0) {
			return DIR_NONE;
		}

		if (line.charAt (0) != '#') {
			return DIR_NONE;
		}

		int type, pos;
		StringHolder word = new StringHolder (line);
		pos = getWord (word, 1);

		// determine directive type
		if (word.length () == 0) {
			return DIR_INVALID;
		} else if (word.equals ("ifdef")) {
			type = DIR_IFDEF;
		} else if (word.equals ("ifndef")) {
			type = DIR_IFNDEF;
		} else if (word.equals ("elifdef")) {
			type = DIR_ELIFDEF;
		} else if (word.equals ("else")) {
			return DIR_ELSE;
		} else if (word.equals ("endif")) {
			return DIR_ENDIF;
		} else if (word.equals ("error")) {
			line = line.substring (pos);
			return DIR_ERROR;
		} else {
			return DIR_INVALID;
		}

		// get the symbol
		getWord (line, pos + 1);
		return type;
	} // checkDirective (String)


	public final void fatal (String msg) {
		System.err.println (msg);
		System.exit (1);
	} // fatal (String)

} // final class Preprocessor


/*******************************************************************\
| StringHolder                                                      |
\*******************************************************************/

final class StringHolder {
	public String string;

	public StringHolder (String value) {
		string = value;
	}

	public StringHolder (StringHolder value) {
		string = value.string;
	}

	public final boolean equals (String value) {
		return string.equals (value);
	}

	public final int length () {
		return string.length ();
	}

	public final char charAt (int pos) {
		return string.charAt (pos);
	}

	public final StringHolder substring (int beg) {
		int len = string.length ();
		beg = (beg >= len) ? len - 1 : beg;
		string = string.substring (beg);
		return this;
	}

	public final StringHolder substring (int beg, int end) {
		int len = string.length ();
		beg = (beg >= len) ? len - 1 : beg;
		end = (end >= len) ? len : end;
		string = string.substring (beg, end);
		return this;
	}

} // final class StringHolder


